# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * Copyright (c) 2011, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""
Locator service uses transport layer to search
for peers and to collect and maintain up-to-date
data about peer's attributes.
"""

import platform
import threading
import time
import socket

from .. import locator
from ...util import logging
from ...channel import fromJSONSequence, toJSONSequence, STATE_CLOSED
from ...channel.ChannelProxy import ChannelProxy
from ... import compat, protocol, services, channel, peer, errors, transport, process_param_str

# Flag indicating whether tracing of the the discovery activity is enabled.
__TRACE_DISCOVERY__ = False


class SubNet(object):
    def __init__(self, prefix_length, address, broadcast):
        self.prefix_length = prefix_length
        self.address = address
        self.broadcast = broadcast
        self.last_slaves_req_time = 0

    def contains(self, addr):
        if addr is None or self.address is None:
            return False
        a1 = addr.getAddress()
        a2 = self.address.getAddress()
        if len(a1) != len(a2):
            return False
        i = 0
        if self.prefix_length <= len(a1) * 8:
            l = self.prefix_length
        else:
            l = len(a1) * 8
        while i + 8 <= l:
            n = int(i / 8)
            if a1[n] != a2[n]:
                return False
            i += 8
        while i < l:
            n = int(i / 8)
            m = 1 << (7 - i % 8)
            if (a1[n] & m) != (a2[n] & m):
                return False
            i += 1
        return True

    def __eq__(self, o):
        if not isinstance(o, SubNet):
            return False
        return self.prefix_length == o.prefix_length and \
            self.broadcast == o.broadcast and \
            self.address == o.address

    def __hash__(self):
        return hash(self.address)

    def __str__(self):
        return "%s/%d" % (self.address.getHostAddress(), self.prefix_length)


class Slave(object):
    # Time of last packet receiver from self slave
    last_packet_time = 0
    # Time of last REQ_SLAVES packet received from self slave
    last_req_slaves_time = 0

    def __init__(self, address, port):
        self.address = address
        self.port = port

    def __str__(self):
        return "%s/%d" % (self.address.getHostAddress(), self.port)


class AddressCacheItem(object):
    address = None
    time_stamp = 0
    used = False

    def __init__(self, host):
        self.host = host


class InetAddress(object):
    "Mimicking Java InetAddress class"
    def __init__(self, host, addr):
        self.host = host
        self.addr = addr

    def getAddress(self):
        return socket.inet_aton(self.addr)

    def getHostAddress(self):
        return self.addr

    def __eq__(self, other):
        if not isinstance(other, InetAddress):
            return False
        return self.addr == other.addr

    def __str__(self):
        return "%s/%s" % (self.host or "", self.addr)

    def __hash__(self):
        return hash(self.addr)


class InputPacket(object):
    "Wrapper for UDP packet data."
    def __init__(self, data, addr, port):
        self.data = data
        self.addr = addr
        self.port = port

    def getLength(self):
        return len(self.data)

    def getData(self):
        return self.data

    def getPort(self):
        return self.port

    def getAddress(self):
        return self.addr

    def __str__(self):
        return "[address=%s,port=%d,data=\"%s\"]" % \
               (self.getAddress(), self.getPort(), self.data)


def is_matching_peer(peer_info, remote_peer):
    attrs = remote_peer.getAttributes()
    if peer.ATTR_ID in peer_info:
        return attrs[peer.ATTR_ID] == peer_info[peer.ATTR_ID]
    return attrs[peer.ATTR_TRANSPORT_NAME] == peer_info[peer.ATTR_TRANSPORT_NAME] and \
           attrs[peer.ATTR_IP_HOST] == peer_info[peer.ATTR_IP_HOST] and \
           attrs[peer.ATTR_IP_PORT] == peer_info[peer.ATTR_IP_PORT]


def fill_peer_attributes(peer_info):
    if peer.ATTR_ID not in peer_info:
        peer_info[peer.ATTR_ID] = "{}:{}:{}".format(
            peer_info[peer.ATTR_TRANSPORT_NAME],
            peer_info[peer.ATTR_IP_HOST],
            peer_info[peer.ATTR_IP_PORT]
        )
    else:
        if peer.ATTR_TRANSPORT_NAME not in peer_info or \
           peer.ATTR_IP_PORT not in peer_info or \
           peer.ATTR_IP_HOST not in peer_info:
            peer_info[peer.ATTR_ID] = process_param_str(peer_info[peer.ATTR_ID])
            _id = peer_info[peer.ATTR_ID]
            peer_info[peer.ATTR_TRANSPORT_NAME], peer_info[peer.ATTR_IP_HOST], peer_info[peer.ATTR_IP_PORT] = \
                _id.split(":")


DISCOVEY_PORT = 1534
MAX_PACKET_SIZE = 9000 - 40 - 8
PREF_PACKET_SIZE = 1500 - 40 - 8

# TODO: research usage of DNS-SD (DNS Service Discovery) to discover TCF peers


class LocatorService(locator.LocatorService):
    locator = None
    peers = {}  # str->Peer
    listeners = []  # list of LocatorListener
    error_log = set()  # set of str
    addr_list = []
    addr_cache = {}  # str->AddressCacheItem
    addr_request = False
    local_peer = None
    last_master_packet_time = 0

    @classmethod
    def getLocalPeer(cls):
        return cls.local_peer

    @classmethod
    def getListeners(cls):
        return cls.listeners[:]

    @classmethod
    def startup(cls):
        if cls.locator:
            cls.locator._startup()

    @classmethod
    def shutdown(cls):
        if cls.locator:
            cls.locator._shutdown()

    def __init__(self):
        self._error_log_lock = threading.RLock()
        self._alive = False
        LocatorService.locator = self
        LocatorService.local_peer = peer.LocalPeer()

    def _startup(self):
        if self._alive:
            return
        self._alive = True
        self._addr_cache_lock = threading.Condition()
        self.subnets = set()
        self.slaves = []
        self.inp_buf = bytearray(MAX_PACKET_SIZE)
        self.out_buf = bytearray(MAX_PACKET_SIZE)
        service = self

        class TimerThread(threading.Thread):
            def __init__(self, _callable):
                self._callable = _callable
                super(TimerThread, self).__init__()

            def run(self):
                while service._alive:
                    try:
                        time.sleep(locator.DATA_RETENTION_PERIOD / 4 / 1000.)
                        protocol.invokeAndWait(self._callable)
                    except RuntimeError:
                        # TCF event dispatch is shut down
                        return
                    except Exception as x:
                        service._log("Unhandled exception in TCF discovery " +
                                     "timer thread", x)
        self.timer_thread = TimerThread(self.__refresh_timer)

        class DNSLookupThread(threading.Thread):
            def run(self):

                while service._alive:
                    try:
                        itemSet = None
                        with service._addr_cache_lock:
                            period = locator.DATA_RETENTION_PERIOD
                            if not LocatorService.addr_request:
                                service._addr_cache_lock.wait(period)
                            msec = int(time.time() * 1000)
                            items = list(LocatorService.addr_cache.items())
                            for host, a in items:
                                if a.time_stamp + period * 10 < msec:
                                    if a.used:
                                        if itemSet is None:
                                            itemSet = set()
                                        itemSet.add(a)
                                    else:
                                        del LocatorService.addr_cache[host]
                            LocatorService.addr_request = False
                        if itemSet is not None:
                            for a in itemSet:
                                addr = None
                                try:
                                    addr = socket.gethostbyname(a.host)
                                except socket.gaierror:
                                    pass
                                with service._addr_cache_lock:
                                    if addr is None:
                                        a.address = None
                                    else:
                                        a.address = InetAddress(a.host, addr)
                                    a.time_stamp = msec
                                    a.used = False
                    except Exception as x:
                        service._log("Unhandled exception in TCF discovery " +
                                     "DNS lookup thread", x)
        self.dns_lookup_thread = DNSLookupThread()

        class InputThread(threading.Thread):
            def __init__(self, _callable):
                self._callable = _callable
                super(InputThread, self).__init__()

            def run(self):
                try:
                    while service._alive:
                        sock = service.socket
                        try:
                            data, addr = sock.recvfrom(MAX_PACKET_SIZE)
                            if addr and len(addr) >= 2:
                                p = InputPacket(data,
                                                InetAddress(None, addr[0]),
                                                addr[1])
                                protocol.invokeAndWait(self._callable, p)
                        except RuntimeError:
                            # TCF event dispatch is shutdown
                            return
                        except socket.error as x:
                            if sock != service.socket:
                                continue
                            # frequent  error on windows, unknown reason
                            if x.errno == 10054:
                                continue
                            port = sock.getsockname()[1]
                            service._log("Cannot read from datagram socket " +
                                         "at port %d" % port, x)
                            time.sleep(2)
                except Exception as x:
                    service._log("Unhandled exception in socket reading " +
                                 "thread", x)
        self.input_thread = InputThread(self.__handleDatagramPacket)
        try:
            self.loopback_addr = InetAddress(None, "127.0.0.1")
            tcfversion = 'TCF%s\0\0\0\0' % locator.CONF_VERSION
            self.out_buf[0:8] = [ord(c) for c in tcfversion]
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            try:
                self.socket.bind(('', DISCOVEY_PORT))
                if __TRACE_DISCOVERY__:
                    logging.trace("Became the master agent (bound to port " +
                                  "%d)" % self.socket.getsockname()[1])
            except socket.error as x:
                self.socket.bind(('', 0))
                if __TRACE_DISCOVERY__:
                    logging.trace("Became a slave agent (bound to port " +
                                  "%d)" % self.socket.getsockname()[1])
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            self.input_thread.setName("TCF Locator Receiver")
            self.timer_thread.setName("TCF Locator Timer")
            self.dns_lookup_thread.setName("TCF Locator DNS Lookup")
            self.input_thread.setDaemon(True)
            self.timer_thread.setDaemon(True)
            self.dns_lookup_thread.setDaemon(True)
            self.input_thread.start()
            self.timer_thread.start()
            self.dns_lookup_thread.start()

            class LocatorListener(locator.LocatorListener):
                def peerAdded(self, peer):
                    service._sendPeerInfo(peer, None, 0)

                def peerChanged(self, peer):
                    service._sendPeerInfo(peer, None, 0)

            self.listeners.append(LocatorListener())
            self.__refreshSubNetList()
            self.__sendPeersRequest(None, 0)
            self.__sendAll(None, 0, None, int(time.time() * 1000))
        except Exception as x:
            self._log("Cannot open UDP socket for TCF discovery protocol", x)

    def _shutdown(self):
        if self._alive:
            self._alive = False
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
            except Exception:
                pass
            try:
                self.socket.close()
            except Exception:
                pass
            try:
                with self._addr_cache_lock:
                    self._addr_cache_lock.notify_all()
            except Exception:
                pass

    def __makeErrorReport(self, code, msg):
        err = {}
        err[errors.ERROR_TIME] = int(time.time() * 1000)
        err[errors.ERROR_CODE] = code
        err[errors.ERROR_FORMAT] = msg
        return err

    def _openPeer(self, reply_channel, token, remote_peer):
        c = remote_peer.openChannel()

        class ChannelListener(channel.ChannelListener):
            def onChannelOpened(self):
                c.removeChannelListener(self)
                c.add_linked_channel(reply_channel)
                reply_channel.sendResult(token, toJSONSequence((None, c.remote_peer.getAttributes())))

            def onChannelClosed(self, error):
                reply_channel.sendResult(token, toJSONSequence((str(error), None)))

        c.addChannelListener(ChannelListener())

    def _command(self, channel, token, name, data):
        try:
            if name == "redirect":
                peer_id = fromJSONSequence(data)[0]
                _peer = self.peers.get(peer_id)
                if _peer is None:
                    errNum = errors.TCF_ERROR_UNKNOWN_PEER
                    error = self.__makeErrorReport(errNum, "Unknown peer ID")
                    channel.sendResult(token, toJSONSequence((error,)))
                    return
                channel.sendResult(token, toJSONSequence((None,)))
                if isinstance(_peer, peer.LocalPeer):
                    seq = (channel.get_local_services(),)
                    channel.sendEvent(protocol.getLocator(), "Hello",
                                      toJSONSequence(seq))
                    return
                ChannelProxy(channel, _peer.openChannel())
            elif name == "sync":
                channel.sendResult(token, None)
            elif name == "getPeers":
                arr = []
                for p in list(self.peers.values()):
                    arr.append(p.getAttributes())
                channel.sendResult(token, toJSONSequence((None, arr)))
            elif name == "getChannels":
                channels = transport.getOpenChannels()
                arr = [c.remote_peer.getAttributes() for c in channels]
                channel.sendResult(token, toJSONSequence((None, arr)))
            elif name == "connectPeer":
                peer_info = fromJSONSequence(data)[0]
                fill_peer_attributes(peer_info)
                chan = None
                err = None
                # check channel does not already exist
                channels = transport.getOpenChannels()
                for c in channels:
                    if is_matching_peer(peer_info, c.remote_peer):
                        chan = c
                        break

                if not chan:
                    remote_peer = None
                    # check peer already exists
                    for p in self.peers.values():
                        if is_matching_peer(peer_info, p):
                            remote_peer = p
                            break
                    if not remote_peer:
                        remote_peer = peer.RemotePeer(peer_info)
                    # connect peer
                    if remote_peer:
                        self._openPeer(channel, token, remote_peer)
                else:
                    chan.add_linked_channel(channel)
                    if err:
                        channel.sendResult(token, toJSONSequence((err, None)))
                    else:
                        channel.sendResult(token, toJSONSequence((None, chan.remote_peer.getAttributes())))

            elif name == "disconnectPeer":
                peer_info = fromJSONSequence(data)[0]
                chan = None
                err = None
                # check channel does not already exist
                channels = transport.getOpenChannels()
                for c in channels:
                    if is_matching_peer(peer_info, c.remote_peer):
                        chan = c
                        break

                if chan:
                    chan.close()
                else:
                    err = "Could not find channel to disconnect."

                if err:
                    channel.sendResult(token, toJSONSequence((err, None)))
                else:
                    channel.sendResult(token, toJSONSequence((None, chan.getState() == STATE_CLOSED)))

            else:
                channel.rejectCommand(token)
        except Exception as x:
            channel.terminate(x)

    def _log(self, msg, x):
        if not self._alive:
            return
        # Don't report same error multiple times to avoid filling up the log
        # file.
        try:
            with self._error_log_lock:
                if msg in self.error_log:
                    return
                self.error_log.add(msg)
        except TypeError:
            # If the error_log_lock thread is dead, it just means that we are
            # shutting down. The _alive value does not seem to be up to date
            # in some cases ...
            return
        protocol.log(msg, x)

    def __getInetAddress(self, host):
        if not host:
            return None
        with self._addr_cache_lock:
            i = self.addr_cache.get(host)
            if i is None:
                i = AddressCacheItem(host)
                ch = host[0]
                if ch == '[' or ch == ':' or ch >= '0' and ch <= '9':
                    try:
                        addr = socket.gethostbyname(host)
                        i.address = InetAddress(host, addr)
                    except socket.gaierror:
                        pass
                    i.time_stamp = int(time.time() * 1000)
                else:
                    # socket.gethostbyname() can cause long delay - delegate to
                    # background thread
                    LocatorService.addr_request = True
                    self._addr_cache_lock.notify()
                self.addr_cache[host] = i
            i.used = True
            return i.address

    def __refresh_timer(self):
        tm = int(time.time() * 1000)
        # Cleanup slave table
        if self.slaves:
            i = 0
            while i < len(self.slaves):
                s = self.slaves[i]
                if s.last_packet_time + locator.DATA_RETENTION_PERIOD < tm:
                    del self.slaves[i]
                else:
                    i += 1

        # Cleanup peers table
        stale_peers = None
        for p in list(self.peers.values()):
            if isinstance(p, peer.RemotePeer):
                if p.getLastUpdateTime() + locator.DATA_RETENTION_PERIOD < tm:
                    if stale_peers is None:
                        stale_peers = []
                    stale_peers.append(p)
        if stale_peers is not None:
            for p in stale_peers:
                p.dispose()

        # Try to become a master
        try:
            port = self.socket.getsockname()[1]
        except OSError:
            port = ""
        period = int(locator.DATA_RETENTION_PERIOD / 2)
        if port != DISCOVEY_PORT and \
                self.last_master_packet_time + period <= tm:
            s0 = self.socket
            s1 = None
            try:
                s1 = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                s1.bind((socket.gethostname(), DISCOVEY_PORT))
                s1.setsockopt(socket.SOL_UDP, socket.SO_BROADCAST, 1)
                self.socket = s1
                s0.close()
            except:
                pass
        self.__refreshSubNetList()
        if port != DISCOVEY_PORT:
            for subnet in self.subnets:
                self.__addSlave(subnet.address, port, tm, tm)
        self.__sendAll(None, 0, None, tm)

    def __addSlave(self, addr, port, timestamp, time_now):
        for s in self.slaves:
            if s.port == port and s.address == addr:
                if s.last_packet_time < timestamp:
                    s.last_packet_time = timestamp
                return s
        s = Slave(addr, port)
        s.last_packet_time = timestamp
        self.slaves.append(s)
        self.__sendPeersRequest(addr, port)
        self.__sendAll(addr, port, s, time_now)
        self.__sendSlaveInfo(s, time_now)
        return s

    def __refreshSubNetList(self):
        subNetSet = set()
        try:
            self.__getSubNetList(subNetSet)
        except Exception as x:
            self._log("Cannot get list of network interfaces", x)
        for s in tuple(self.subnets):
            if s in subNetSet:
                continue
            self.subnets.remove(s)
        for s in subNetSet:
            if s in self.subnets:
                continue
            self.subnets.add(s)
        if __TRACE_DISCOVERY__:
            buf = "Refreshed subnet list:"
            for subnet in self.subnets:
                buf += "\n\t* address=%s, broadcast=%s" % \
                       (subnet.address, subnet.broadcast)
            logging.trace(buf)

    def __getAllIpAddresses(self):
        import fcntl  # @UnresolvedImport
        import struct
        import array

        nBytes = 8192
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        names = array.array('B', [0] * nBytes)
        ifcfg = struct.unpack(
            'iL', fcntl.ioctl(s.fileno(), 0x8912,
                              struct.pack('iL', nBytes,
                                          names.buffer_info()[0])))[0]

        namestr = names.tostring()
        if namestr and isinstance(namestr[0], int):
            namestr = ''.join(chr(b) for b in namestr)
        res = []

        # the ipconf structure changed at a time, check if there are more than
        # 40 bytes

        ifconfSz = 32
        sz = 32
        altSz = 40

        if len(namestr) > sz:
            # check for name at 32->32+16
            secondName = str(namestr[sz:sz + 16].split('\0', 1)[0])
            secondAltName = str(namestr[altSz:altSz + 16].split('\0', 1)[0])

            if (not secondName.isalnum()) and (secondAltName.isalnum()):
                ifconfSz = 40

        for ix in range(0, ifcfg, ifconfSz):
            ipStartIx = ix + 20
            ipEndIx = ix + 24
            ip = namestr[ipStartIx:ipEndIx]
            res.append(str(ord(ip[0])) + '.' + str(ord(ip[1])) + '.' +
                       str(ord(ip[2])) + '.' + str(ord(ip[3])))

        return (res)

    def __getSubNetList(self, _set):
        hostname = socket.gethostname()
        if len(self.addr_list) == 0:
            # Create the list of IP address for this host
            _, _, self.addr_list = socket.gethostbyname_ex(hostname)
            if "127.0.0.1" not in self.addr_list:
                self.addr_list.append("127.0.0.1")

            # On unix hosts, use sockets to get the other interfaces IPs

            if (platform.system() != 'Windows'):
                for ip_addr in self.__getAllIpAddresses():
                    if ip_addr not in self.addr_list:
                        self.addr_list.append(ip_addr)

        for address in self.addr_list:
            rawaddr = socket.inet_aton(address)
            if len(rawaddr) != 4:
                continue
            if isinstance(rawaddr, str):
                rawaddr = rawaddr[:3] + '\xFF'
            elif isinstance(rawaddr, bytes):
                rawaddr = bytes([b for b in rawaddr[:3]] + [255])
            broadcast = socket.inet_ntoa(rawaddr)
            _set.add(SubNet(24, InetAddress(hostname, address),
                            InetAddress(None, broadcast)))

    def __getUTF8Bytes(self, s):
        return s.encode("UTF-8")

    # Used for tracing
    packetTypes = [
        None,
        "CONF_REQ_INFO",
        "CONF_PEER_INFO",
        "CONF_REQ_SLAVES",
        "CONF_SLAVES_INFO",
        "CONF_PEERS_REMOVED"
    ]

    def __sendDatagramPacket(self, subnet, size, addr, port):
        try:
            if addr is None:
                addr = subnet.broadcast
                port = DISCOVEY_PORT
                for slave in self.slaves:
                    self.__sendDatagramPacket(subnet, size, slave.address,
                                              slave.port)
            if not subnet.contains(addr):
                return False
            if port == self.socket.getsockname()[1] and \
               addr == subnet.address:
                return False

            self.socket.sendto(compat.str2bytes(self.out_buf[:size]),
                               (addr.getHostAddress(), port))

            if __TRACE_DISCOVERY__:
                attrs = None
                if self.out_buf[4] == locator.CONF_PEER_INFO:
                    attrs = self.__parsePeerAttributes(self.out_buf, 8)
                elif self.out_buf[4] == locator.CONF_SLAVES_INFO:
                    attrs = self.__parseIDs(self.out_buf, size)
                elif self.out_buf[4] == locator.CONF_PEERS_REMOVED:
                    attrs = self.__parseIDs(self.out_buf, size)
                self.__traceDiscoveryPacket(False,
                                            self.packetTypes[self.out_buf[4]],
                                            attrs, addr, port)
        except Exception as x:
            self._log("Cannot send datagram packet to %s" % addr, x)
            return False
        return True

    def __parsePeerAttributes(self, data, size):
        """
        Parse peer attributes in CONF_PEER_INFO packet data

        @param data - the packet data
        @param size - the packet size
        @return a map containing the attributes
        """
        attrs = {}
        # Remove packet header
        s = data[8:size].decode("UTF-8")
        l = len(s)
        i = 0
        while i < l:
            i0 = i
            while i < l and s[i] != '=' and s[i] != '\0':
                i += 1
            i1 = i
            if i < l and s[i] == '=':
                i += 1
            i2 = i
            while i < l and s[i] != '\0':
                i += 1
            i3 = i
            if i < l and s[i] == '\0':
                i += 1
            key = s[i0:i1]
            val = s[i2:i3]
            attrs[key] = val
        return attrs

    def __parseIDs(self, data, size):
        """
        Parse list of IDs in CONF_SLAVES_INFO and CONF_PEERS_REMOVED packet
        data.

        @param data - the packet data
        @param size - the packet size
        @return a map containing the IDs
        """
        cnt = 0
        attrs = {}
        s = data[8:].decode("UTF-8")
        l = len(s)
        i = 0
        while i < l:
            i0 = i
            while i < l and s[i] != '\0':
                i += 1
            if i > i0:
                _id = s[i0:i]
                attrs[str(cnt)] = _id
                cnt += 1
            while i < l and s[i] == '\0':
                i += 1
        return attrs

    def __sendPeersRequest(self, addr, port):
        self.out_buf[4] = locator.CONF_REQ_INFO
        for subnet in self.subnets:
            self.__sendDatagramPacket(subnet, 8, addr, port)

    def _sendPeerInfo(self, _peer, addr, port):
        attrs = _peer.getAttributes()
        peer_addr = self.__getInetAddress(attrs.get(peer.ATTR_IP_HOST))
        if peer_addr is None:
            return
        if attrs.get(peer.ATTR_IP_PORT) is None:
            return
        self.out_buf[4] = locator.CONF_PEER_INFO
        i = 8

        for subnet in self.subnets:
            if isinstance(_peer, peer.RemotePeer):
                try:
                    if self.socket.getsockname()[1] != DISCOVEY_PORT:
                        return
                except:
                    pass
                if not subnet.address == self.loopback_addr and \
                   not subnet.address == peer_addr:
                    continue
            if not subnet.address == self.loopback_addr:
                if not subnet.contains(peer_addr):
                    continue
            if i == 8:
                sb = []
                for key in list(attrs.keys()):
                    sb.append(str(key) + '=' + str(attrs.get(key)))
                bt = self.__getUTF8Bytes('\0'.join(sb))
                if i + len(bt) > len(self.out_buf):
                    return
                self.out_buf[i:i + len(bt)] = bt
                i += len(bt)
            if self.__sendDatagramPacket(subnet, i, addr, port):
                subnet.send_all_ok = True

    def __sendEmptyPacket(self, addr, port):
        self.out_buf[4] = locator.CONF_SLAVES_INFO
        for subnet in self.subnets:
            if subnet.send_all_ok:
                continue
            self.__sendDatagramPacket(subnet, 8, addr, port)

    def __sendAll(self, addr, port, sl, tm):
        for subnet in self.subnets:
            subnet.send_all_ok = False
        for peer in list(self.peers.values()):
            self._sendPeerInfo(peer, addr, port)
        if addr is not None and sl is not None and \
           sl.last_req_slaves_time + locator.DATA_RETENTION_PERIOD >= tm:
            self.__sendSlavesInfo(addr, port, tm)
        self.__sendEmptyPacket(addr, port)

    def __sendSlavesRequest(self, subnet, addr, port):
        self.out_buf[4] = locator.CONF_REQ_SLAVES
        self.__sendDatagramPacket(subnet, 8, addr, port)

    def __sendSlaveInfo(self, x, tm):
        ttl = x.last_packet_time + locator.DATA_RETENTION_PERIOD - tm
        if ttl <= 0:
            return
        self.out_buf[4] = locator.CONF_SLAVES_INFO
        for subnet in self.subnets:
            if not subnet.contains(x.address):
                continue
            i = 8
            s = "%d:%s:%s" % (ttl, x.port, x.address.getHostAddress())
            bt = self.__getUTF8Bytes(s)
            self.out_buf[i:i + len(bt)] = bt
            i += len(bt)
            self.out_buf[i] = 0
            i += 1
            for y in self.slaves:
                if not subnet.contains(y.address):
                    continue
                if y.last_req_slaves_time + locator.DATA_RETENTION_PERIOD < tm:
                    continue
                self.__sendDatagramPacket(subnet, i, y.address, y.port)

    def __sendSlavesInfo(self, addr, port, tm):
        self.out_buf[4] = locator.CONF_SLAVES_INFO
        for subnet in self.subnets:
            if not subnet.contains(addr):
                continue
            i = 8
            for x in self.slaves:
                ttl = x.last_packet_time + locator.DATA_RETENTION_PERIOD - tm
                if ttl <= 0:
                    return
                if x.port == port and x.address == addr:
                    continue
                if not subnet.address == self.loopback_addr:
                    if not subnet.contains(x.address):
                        continue
                subnet.send_all_ok = True
                s = "%d:%s:%s" % \
                    (x.last_packet_time, x.port, x.address.getHostAddress())
                bt = self.__getUTF8Bytes(s)
                if i > 8 and i + len(bt) >= PREF_PACKET_SIZE:
                    self.__sendDatagramPacket(subnet, i, addr, port)
                    i = 8
                self.out_buf[i:len(bt)] = bt
                i += len(bt)
                self.out_buf[i] = 0
                i += 1
            if i > 8:
                self.__sendDatagramPacket(subnet, i, addr, port)

    def __isRemote(self, address, port):
        if port != self.socket.getsockname()[1]:
            return True
        for s in self.subnets:
            if s.address == address:
                return False
        return True

    def __handleDatagramPacket(self, p):
        try:
            tm = int(time.time() * 1000)
            buf = p.getData().decode("UTF-8")
            length = p.getLength()
            if length < 8:
                return
            if buf[0] != 'T':
                return
            if buf[1] != 'C':
                return
            if buf[2] != 'F':
                return
            if buf[3] != locator.CONF_VERSION:
                return
            remote_port = p.getPort()
            remote_address = p.getAddress()
            if self.__isRemote(remote_address, remote_port):
                if buf[4] == locator.CONF_PEERS_REMOVED:
                    self.__handlePeerRemovedPacket(p)
                else:
                    sl = None
                    if remote_port != DISCOVEY_PORT:
                        sl = self.__addSlave(remote_address, remote_port, tm,
                                             tm)
                    code = ord(buf[4])
                    if code == locator.CONF_PEER_INFO:
                        self.__handlePeerInfoPacket(p)
                    elif code == locator.CONF_REQ_INFO:
                        self.__handleReqInfoPacket(p, sl, tm)
                    elif code == locator.CONF_SLAVES_INFO:
                        self.__handleSlavesInfoPacket(p, tm)
                    elif code == locator.CONF_REQ_SLAVES:
                        self.__handleReqSlavesPacket(p, sl, tm)
                    for subnet in self.subnets:
                        if not subnet.contains(remote_address):
                            continue
                        delay = int(locator.DATA_RETENTION_PERIOD / 3)
                        if remote_port != DISCOVEY_PORT:
                            delay = int(locator.DATA_RETENTION_PERIOD / 32)
                        elif subnet.address != remote_address:
                            delay = int(locator.DATA_RETENTION_PERIOD / 2)
                        if subnet.last_slaves_req_time + delay <= tm:
                            self.__sendSlavesRequest(subnet, remote_address,
                                                     remote_port)
                            subnet.last_slaves_req_time = tm
                        if subnet.address == remote_address and \
                           remote_port == DISCOVEY_PORT:
                            self.last_master_packet_time = tm
        except Exception as x:
            self._log("Invalid datagram packet received from %s/%s" %
                      (p.getAddress(), p.getPort()), x)

    def __handlePeerInfoPacket(self, p):
        try:
            attrs = self.__parsePeerAttributes(p.getData(), p.getLength())
            if __TRACE_DISCOVERY__:
                self.__traceDiscoveryPacket(True, "CONF_PEER_INFO", attrs, p)
            _id = attrs.get(peer.ATTR_ID)
            if _id is None:
                raise RuntimeError("Invalid peer info: no ID")
            ok = True
            host = attrs.get(peer.ATTR_IP_HOST)
            if host is not None:
                ok = False
                peer_addr = self.__getInetAddress(host)
                if peer_addr is not None:
                    for subnet in self.subnets:
                        if subnet.contains(peer_addr):
                            ok = True
                            break
            if ok:
                _peer = self.peers.get(_id)
                if isinstance(_peer, peer.RemotePeer):
                    _peer.updateAttributes(attrs)
                elif _peer is None:
                    peer.RemotePeer(attrs)
        except Exception as x:
            self._log("Invalid datagram packet received from %s/%s" %
                      (p.getAddress(), p.getPort()), x)

    def __handleReqInfoPacket(self, p, sl, tm):
        if __TRACE_DISCOVERY__:
            self.__traceDiscoveryPacket(True, "CONF_REQ_INFO", None, p)
        self.__sendAll(p.getAddress(), p.getPort(), sl, tm)

    def __handleSlavesInfoPacket(self, p, time_now):
        try:
            attrs = self.__parseIDs(p.getData(), p.getLength())
            if __TRACE_DISCOVERY__:
                self.__traceDiscoveryPacket(True, "CONF_SLAVES_INFO", attrs, p)
            for s in list(attrs.values()):
                i = 0
                l = len(s)
                time0 = i
                while i < l and s[i] != ':' and s[i] != '\0':
                    i += 1
                time1 = i
                if i < l and s[i] == ':':
                    i += 1
                port0 = i
                while i < l and s[i] != ':' and s[i] != '\0':
                    i += 1
                port1 = i
                if i < l and s[i] == ':':
                    i += 1
                host0 = i
                while i < l and s[i] != '\0':
                    i += 1
                host1 = i
                port = int(s[port0:port1])
                timestamp = s[time0:time1]
                host = s[host0:host1]
                if port != DISCOVEY_PORT:
                    addr = self.__getInetAddress(host)
                    if addr is not None:
                        delta = 10006030  # 30 minutes
                        if len(timestamp) > 0:
                            time_val = int(timestamp)
                        else:
                            time_val = time_now
                        if time_val < 3600000:
                            # Time stamp is "time to live" in milliseconds
                            time_val = time_now + int(time_val / 1000) - \
                                locator.DATA_RETENTION_PERIOD
                        elif time_val < int(time_now / 1000) + 50000000:
                            # Time stamp is in seconds
                            time_val = 1000
                        else:
                            # Time stamp is in milliseconds
                            pass
                        if time_val < time_now - delta or \
                           time_val > time_now + delta:
                            msg = "Invalid slave info timestamp: %s -> %s" % (
                                  timestamp,
                                  time.strftime("%Y-%m-%d %H:%M:%S",
                                                time.localtime(time_val /
                                                               1000.)))

                            self._log("Invalid datagram packet received " +
                                      "from %s/%s" % (p.getAddress(),
                                                      p.getPort()),
                                      Exception(msg))
                            time_val = time_now - \
                                int(locator.DATA_RETENTION_PERIOD / 2)
                        self.__addSlave(addr, port, time_val, time_now)
        except Exception as x:
            self._log("Invalid datagram packet received from " +
                      "%s/%s" % (p.getAddress(), p.getPort()), x)

    def __handleReqSlavesPacket(self, p, sl, tm):
        if __TRACE_DISCOVERY__:
            self.__traceDiscoveryPacket(True, "CONF_REQ_SLAVES", None, p)
        if sl is not None:
            sl.last_req_slaves_time = tm
        self.__sendSlavesInfo(p.getAddress(), p.getPort(), tm)

    def __handlePeerRemovedPacket(self, p):
        try:
            attrs = self.__parseIDs(p.getData(), p.getLength())
            if __TRACE_DISCOVERY__:
                self.__traceDiscoveryPacket(True, "CONF_PEERS_REMOVED", attrs,
                                            p)
            for _id in list(attrs.values()):
                _peer = self.peers.get(_id)
                if isinstance(_peer, peer.RemotePeer):
                    _peer.dispose()
        except Exception as x:
            self._log("Invalid datagram packet received from %s/%s" % (
                      p.getAddress(), p.getPort()), x)

    @classmethod
    def getLocator(cls):
        return cls.locator

    def getPeers(self):
        assert protocol.isDispatchThread()
        return self.peers

    def redirect(self, peer, done):
        raise RuntimeError("Channel redirect cannot be done on local peer")

    def sync(self, done):
        raise RuntimeError("Channel sync cannot be done on local peer")

    def addListener(self, listener):
        assert listener is not None
        assert protocol.isDispatchThread()
        self.listeners.append(listener)

    def removeListener(self, listener):
        assert protocol.isDispatchThread()
        self.listeners.remove(listener)

    @classmethod
    def __traceDiscoveryPacket(cls, received, packet_type, attrs, addr,
                               port=None):
        """
        Log that a TCF Discovery packet has be sent or received. The trace is
        sent to stdout. This should be called only if the tracing has been
        turned on.

        @param received
                   True if the packet was sent, otherwise it was received
        @param packet_type
                   a string specifying the type of packet, e.g.,
                   "CONF_PEER_INFO"
        @param attrs
                   a set of attributes relevant to the type of packet
                   (typically a peer's attributes)
        @param addr
                   the network address the packet is being sent to
        @param port
                   the port the packet is being sent to
        """
        assert __TRACE_DISCOVERY__
        if port is None:
            # addr is a InputPacket
            port = addr.getPort()
            addr = addr.getAddress()
        buf = str(packet_type)
        buf += (" sent to ", " received from ")[received]
        buf += "%s/%s" % (addr, port)
        if attrs is not None:
            for key, value in list(attrs.items()):
                buf += "\n\t%s=%s" % (key, value)
        logging.trace(buf)


class LocatorServiceProvider(services.ServiceProvider):
    def get_local_service(self, _channel):
        class CommandServer(channel.CommandServer):
            def command(self, token, name, data):
                LocatorService.locator._command(_channel, token, name, data)
        _channel.addCommandServer(LocatorService.locator, CommandServer())
        return (LocatorService.locator,)

services.add_service_provider(LocatorServiceProvider())



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy import dm
from chipscopy.client.core_property_client import CorePropertyClient
from chipscopy.tcf.services import DoneHWCommand


SERVICE_NAME = "AxisTrace"


class AxisTraceClient(CorePropertyClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.core_type == 5

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(SERVICE_NAME)

    #
    #  General methods
    #
    def initialize(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.initialize(self.ctx, done_cb))

    def terminate(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.terminate(self.ctx, done_cb))



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""Locator service uses transport layer to search for peers and to collect data
about peer's attributes and capabilities (services).

.. |getAgentID| replace:: :meth:`~LocatorService.getAgentID`
.. |redirect| replace:: :meth:`~LocatorService.redirect`
.. |sync| replace:: :meth:`~LocatorService.sync`
.. |DoneGetAgentID| replace:: :class:`DoneGetAgentID`
.. |DoneRedirect| replace:: :class:`DoneRedirect`
.. |DoneSync| replace:: :class:`DoneSync`
.. |LocatorListener| replace:: :class:`LocatorListener`
.. |Peer| replace:: :class:`~tcf.peer.Peer`

Discovery mechanism depends on transport protocol and is part of that protocol
handler. Targets, known to other hosts, can be found through remote instances
of Locator service.

Automatically discovered targets require no further configuration. Additional
targets can be configured manually.

Clients should use :meth:`tcf.protocol.getLocator` to obtain local instance of
locator, then :meth:`LocatorService.getPeers` can be used to get list of
available peers (hosts and targets).

Service Methods
---------------
.. autodata:: NAME
.. autoclass:: LocatorService

addListener
^^^^^^^^^^^
.. automethod:: LocatorService.addListener

getAgentID
^^^^^^^^^^
.. automethod:: LocatorService.getAgentID

getName
^^^^^^^
.. automethod:: LocatorService.getName

getPeers
^^^^^^^^
.. automethod:: LocatorService.getPeers

redirect
^^^^^^^^
.. automethod:: LocatorService.redirect

removeListener
^^^^^^^^^^^^^^
.. automethod:: LocatorService.removeListener

sync
^^^^
.. automethod:: LocatorService.sync

Callback Classes
----------------
DoneGetAgentID
^^^^^^^^^^^^^^
.. autoclass:: DoneGetAgentID
    :members:

DoneRedirect
^^^^^^^^^^^^
.. autoclass:: DoneRedirect
    :members:

DoneSync
^^^^^^^^
.. autoclass:: DoneSync
    :members:

Listener
--------
LocatorListener
^^^^^^^^^^^^^^^
.. autoclass:: LocatorListener
    :members:
"""

from .. import services
from typing import Any, Dict, List

DATA_RETENTION_PERIOD = 60 * 1000
""" Peer data retention period in milliseconds."""
CONF_VERSION = '2'
""" Auto-configuration protocol version."""

# Auto-configuration command and response codes.
CONF_REQ_INFO = 1
CONF_PEER_INFO = 2
CONF_REQ_SLAVES = 3
CONF_SLAVES_INFO = 4
CONF_PEERS_REMOVED = 5

NAME = "Locator"
"""Locator service name."""


class LocatorService(services.Service):
    """TCF Locator service interface."""

    def getAgentID(self, done):
        """Get agent ID of the agent providing the locator service.

        :param done: Call back interface called when operation is completed.
        :type done: |DoneGetAgentID|
        """
        raise NotImplementedError("Abstract method")

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def getPeers(self):
        """Get map (ID -> IPeer) of available peers (hosts and targets).

        The method returns cached (currently known to the framework) list of
        peers. The list is updated according to event received from transport
        layer.

        :returns: A |dict| of Peer ID -> |Peer| association.
        """
        raise NotImplementedError("Abstract method")

    def redirect(self, peer, done):
        """Redirect this service channel to given peer using this service as a
        proxy.

        :param peer: Peer ID or attributes map.
        :type peer: |basestring|
        :param done: Call back interface called when operation is completed.
        :type done: |DoneRedirect|
        """
        raise NotImplementedError("Abstract method")

    def sync(self, done):
        """Call back after TCF messages sent to this target up to this moment
        are delivered.

        This method is intended for synchronization of messages across multiple
        channels.

        .. note:: Cross channel synchronization can reduce performance and
                  throughput. Most clients don't need channel synchronization
                  and should not call this method.

        :param done: Will be executed by dispatch thread after communication
                     messages are delivered to corresponding targets.
        :type done: |DoneSync|

        This is internal API, TCF clients should use module
        :mod:`tcf.protocol`.
        """
        raise NotImplementedError("Abstract method")

    def addListener(self, listener):
        """Add a listener for Locator service events.

        :param listener: Listener to add.
        :type listener: |LocatorListener|
        """
        raise NotImplementedError("Abstract method")

    def removeListener(self, listener):
        """Remove a listener for Locator service events.

        :param listener: Listener to remove.
        :type listener: |LocatorListener|
        """
        raise NotImplementedError("Abstract method")

    def get_remote_peers(self, done: services.DoneHWCommand) -> List[Dict[str, str]]:
        """Get attributes of available peers (hosts and targets) on remote server.

        The method returns cached (currently known to the framework) list of
        peers. The list is updated according to event received from transport
        layer.

        :returns: A |list| of |dict| consisting of |Peer| Attributes.
        """
        raise NotImplementedError("Abstract method")

    def get_remote_channels(self, done: services.DoneHWCommand) -> Dict[str, str]:
        """Get attributes of open channels on a remote server.

        :returns: A |list| of |dict| consisting of |Channel| Attributes.
        """
        raise NotImplementedError("Abstract method")

    def connect_remote_peer(self, peer_attrib: Dict[str, str], done: services.DoneHWCommand) -> Dict[str, str]:
        """Attempts to connect and open a channel with given peer.

        :param peer_attrib: Attributes of the desired peer to which to connect.
        :param done: Callback with result and any error.
        :returns: A |dict| consisting of connected |Peer| Attributes if successful.
        """
        raise NotImplementedError("Abstract method")

    def disconnect_remote_peer(self, peer_attrib: Dict[str, str], done: services.DoneHWCommand) -> bool:
        """Attempts to disconnect and close channel of given peer.

        :param peer_attrib: Attributes of the desired peer to which to disconnect.
        :param done: Callback with result and any error.
        :returns: True if channel is closed.
        """
        raise NotImplementedError("Abstract method")


class DoneGetAgentID(object):
    """Client call back interface for |getAgentID|."""

    def doneGetAgentID(self, token, error, agentID):
        """Called when a peer agent ID request is done.

        :param token: TCF request token corresponding to this done getAgentID
                      call.
        :param error: error description if operation failed, **None** if
                      succeeded.
        :param agentID: Retrieved agent ID.
        :type agentID: |basestring|
        """
        pass


class DoneRedirect(object):
    """Client call back interface for |redirect|."""

    def doneRedirect(self, token, error):
        """Called when a peer redirection request has been done.

        :param token: TCF request token corresponding to this done redirect
                      call.
        :param error: error description if operation failed, **None** if
                      succeeded.
        """
        pass


class DoneSync(object):
    """Client call back interface for |sync|."""
    def doneSync(self, token):
        """Called when a peer sync request has been done.

        :param token: TCF request token corresponding to this done sync call.
        """
        pass


class LocatorListener(object):
    """TCF peer locator listener."""

    def peerAdded(self, peer):
        """Called when a new peer has been added to the known peers.

        :param peer: New peer description.
        :type peer: |Peer|
        """
        pass

    def peerChanged(self, peer):
        """Called when one of the known peers has been modified.

        :param peer: New peer description.
        :type peer: |Peer|
        """
        pass

    def peerRemoved(self, peerID):
        """Called when one of the peers has been shut down.

        :param peerId: Id of the peer which shut down.
        :type peerId: |basestring|
        """
        pass

    def peerHeartBeat(self, peerID):
        """Called when one of the peers says he is alive.

        :param peerId: Id of the peer which sent a heartbeat
        :type peerId: |basestring|
        """
        pass



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from collections.abc import Sequence
from typing import Tuple
from chipscopy import dm
from chipscopy.dm import request
from chipscopy.tcf import services
from chipscopy.tcf.services import local, from_xargs


class CsRequestDone(request.DoneRequest):
    def __init__(self, req, service_name: str, name: str = ""):
        self._name = name

        self.req = req
        self.service_name = service_name

    def done_request(self, token, error, args):
        if error:
            e_msg = f"ChipScope Service {self.service_name} {self._name}: {error}"
            self.req.sendError(e_msg)
        else:
            self.req.sendXicomResult(args)


class CsRequestProgressUpdate(request.ProgressRequest):
    def __init__(self, req, service_name: str, name: str = ""):
        self._name = name

        self.req = req
        self.service_name = service_name

    def done_progress(self, token, error, result):
        self.req.sendXicomProgress(result)


class DefaultCoreService(services.Service):
    """Abstract class for property based debug cores including harden blocks."""

    def getName(self):
        raise NotImplementedError("Abstract method")

    def __init__(self, cls_type: type, progress_update_supported: bool = False):
        super().__init__(None)
        self._class_type = cls_type
        self._progress_update_supported: bool = progress_update_supported

    def _get_args(
        self, req: local.CommandRequest, in_arg_count: int, function_name: str = ""
    ) -> Tuple:
        args = from_xargs(req.args)
        if len(args) < in_arg_count:
            raise Exception(
                f"Insufficient arguments. Need {in_arg_count} argument(s),"
                f" but given only {len(args)} argument(s)."
            )
        csm_name, node = dm.parse_node_id(args[0])
        cs_manager = dm.get_manager(csm_name)
        cs_req = request.CsRequest(
            cs_manager,
            node,
            self._class_type,
            CsRequestDone(req, self.getName(), function_name),
            # Note - Only if the core supports progress_update callbacks create
            #  a CsRequestProgressUpdate object. The "run" func in request.py makes use of
            #  this to decide whether or not to pass this argument to the function being called
            CsRequestProgressUpdate(req, self.getName(), function_name)
            if self._progress_update_supported
            else None,
        )
        return (cs_req, *args[1:])

    def _require_dict(self, req: local.CommandRequest, arg_index: int) -> None:
        if len(req.args) < arg_index + 1 or not isinstance(req.args[arg_index], dict):
            raise ValueError(
                f"Expected argument in position {arg_index + 1} to be a dict or a map."
            )
        return

    def _require_seq(self, req: local.CommandRequest, arg_index: int) -> None:
        if len(req.args) < arg_index + 1 or not isinstance(
            req.args[arg_index], Sequence
        ):
            raise ValueError(
                f"Expected argument in position {arg_index + 1} to be a list."
            )
        return

    def _require_dict_dict(self, req: local.CommandRequest, arg_index: int) -> None:
        self._require_dict(req, arg_index)
        for dd in req.args[arg_index].values():
            if not isinstance(dd, dict):
                raise ValueError(
                    f"Expected argument in position {arg_index + 1} to be a dict-of-dicts or a map-of-maps."
                )
        return

    def _require_seq_dict(self, req: local.CommandRequest, arg_index: int) -> None:
        self._require_seq(req, arg_index)
        for dd in req.args[arg_index]:
            if not isinstance(dd, dict):
                raise ValueError(
                    f"Expected argument in position {arg_index + 1} to be a list-of-dicts."
                )
        return

    def getPropertyCmd(self, req: local.CommandRequest):
        cs_req, property_names = self._get_args(req, 2, "get_property")
        cs_req.get_property(property_names)

    def getPropertyGroupCmd(self, req: local.CommandRequest):
        cs_req, property_group_names = self._get_args(req, 2, "get_property_group")
        cs_req.get_property_group(property_group_names)

    def reportPropertyCmd(self, req: local.CommandRequest):
        cs_req, property_names = self._get_args(req, 2, "report_property")
        cs_req.report_property(property_names)

    def reportPropertyGroupCmd(self, req: local.CommandRequest):
        cs_req, property_group_names = self._get_args(req, 2, "report_property_group")
        cs_req.report_property_group(property_group_names)

    def resetPropertyCmd(self, req: local.CommandRequest):
        cs_req, property_names = self._get_args(req, 2, "reset_property")
        cs_req.reset_property(property_names)

    def resetPropertyGroupCmd(self, req: local.CommandRequest):
        cs_req, property_group_names = self._get_args(req, 2, "reset_property_group")
        cs_req.reset_property_group(property_group_names)

    def setPropertyCmd(self, req: local.CommandRequest):
        cs_req, property_values = self._get_args(req, 2, "set_property")
        self._require_dict(req, 1)
        cs_req.set_property(property_values)

    def setPropertyGroupCmd(self, req: local.CommandRequest):
        cs_req, property_values = self._get_args(req, 2, "set_property_group")
        self._require_dict(req, 1)
        cs_req.set_property_group(property_values)

    def listPropertyGroupsCmd(self, req: local.CommandRequest):
        (cs_req,) = self._get_args(req, 1, "list_property_groups")
        cs_req.list_property_groups()

    def commitPropertyGroupCmd(self, req: local.CommandRequest):
        cs_req, property_group_names = self._get_args(req, 2, "commit_property_group")
        cs_req.commit_property_group(property_group_names)

    def refreshPropertyGroupCmd(self, req: local.CommandRequest):
        cs_req, groups = self._get_args(req, 2, "refresh_property_group")
        cs_req.refresh_property_group(groups)

    def add_to_property_watchlist_cmd(self, req: local.CommandRequest):
        cs_req, options = self._get_args(req, 2, "add_to_property_watchlist")
        cs_req.add_to_watchlist(options)

    def remove_from_property_watchlist_cmd(self, req: local.CommandRequest):
        cs_req, options = self._get_args(req, 2, "remove_from_property_watchlist")
        cs_req.remove_from_watchlist(options)

    # For testing
    def commitMemoryCmd(self, req: local.CommandRequest):
        (cs_req, property_name, buffer, start_byte_index, word_byte_length) = self._get_args(req, 5, "commit_memory")
        return cs_req.commit_memory(property_name, buffer, start_byte_index, word_byte_length)

    # For testing
    def refreshMemoryCmd(self, req: local.CommandRequest):
        (cs_req, property_name, word_count, start_byte_index, word_byte_length) = \
            self._get_args(req, 5, "refresh_memory")
        return cs_req.refresh_memory(property_name, word_count, start_byte_index, word_byte_length)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import re
from chipscopy import dm
from chipscopy.dm import memory, jtag, request
from chipscopy.tcf import protocol
from chipscopy.tcf.services import xicom, ServiceSync
from chipscopy.utils.logger import log
from .view_info import ViewInfo
from .util import sync_call, parse_params, process_param_str
from .util import config


class ServerInfo(dm.Node):
    def __init__(self, channel):
        dm.Node.__init__(self, "")
        self.channel = channel
        self._props_ready = threading.Event()
        self._update_static_info()
        self.views = {}

    def __getattr__(self, attr):
        self._props_ready.wait(5)
        return dm.Node.__getattr__(self, attr)

    def __str__(self):
        self._props_ready.wait(5)
        return "ServerInfo({})".format(self.url)

    @property
    def props(self):
        self._props_ready.wait(5)
        return super().props

    def _update_static_info(self):
        if not protocol.isDispatchThread():
            self._props_ready.clear()
            protocol.invokeLater(self._update_static_info)
            return

        self["services"] = self.channel.getRemoteServices()

        peer = self.channel.getRemotePeer()
        self["url"] = "{}:{}".format(peer.getHost(), peer.getPort())

        proc = self.channel.getRemoteService(xicom.XicomService)

        if proc:

            def done_version_info(token, error, results):
                self._props.update(results)
                self._props_ready.set()

            self.add_pending(proc.get_hw_server_version_info(done_version_info))
        else:
            self._props_ready.set()

    def get_sync_service(self, name):
        return ServiceSync(self.channel, name, True)

    def close(self, wait=True):
        if wait:
            protocol.invokeAndWait(self.channel.close)
        else:
            protocol.invokeLater(self.channel.close)
        self.views.clear()

    @staticmethod
    def _check_thread(command, done, *args, **kwargs):
        if not done:
            return sync_call(command, *args, **kwargs)

        kwargs["done"] = done
        if not protocol.isDispatchThread():
            return protocol.invokeAndWait(command, *args, **kwargs)
        return command(*args, **kwargs)

    def connect_remote(self, url: str, done: request.DoneCallback = None):
        """
        Request that the CS Server connect to another server (usually a hw_server)
        :param url: Url for remote server
        :param done: Done callback (skip if calling synchronously)
        :return: Token if async or Results if sync
        """
        return self._check_thread(self._connect_remote, done, url)

    def _connect_remote(self, url, done):
        locator = self.channel.getRemoteService("Locator")

        def done_remote_connect(token, error, result):
            log.client.info(f"{str(self)} connected to {url}")
            done(token, error, result)

        return locator.connect_remote_peer({"ID": url}, done_remote_connect)

    def disconnect_remote(self, url: str, done: request.DoneCallback = None):
        """
        Request that the CS Server disconnect from another server (usually a hw_server)
        :param url: Url for remote server
        :param done: Done callback (skip if calling synchronously)
        :return: Token if async or Results if sync
        """
        return self._check_thread(self._disconnect_remote, done, url)

    def _disconnect_remote(self, url, done):
        locator = self.channel.getRemoteService("Locator")

        def done_remote_disconnect(token, error, result):
            log.client.info(f"{str(self)} disconnected from {url}")
            done(token, error, result)

        return locator.disconnect_remote_peer({"ID": url}, done_remote_disconnect)

    def connect_xvc(
        self,
        xvc_url: str,
        remote_server: str = "",
        idcode: int = 0,
        done: request.DoneCallback = None,
    ):
        """
        Request remote server with XvcContext service to open an XVC connection with given url
        :param xvc_url: XVC Server url address
        :param remote_server: ID of remote server through which to connect
        :param idcode: Optional parameter to force the XVC server to show a specific idcode
        :param done: Done callback (skip if calling synchronously
        :return: Token if async or Results if sync
        """
        params = parse_params(process_param_str(xvc_url, "2225"))
        return self._check_thread(
            self._connect_xvc, done, params["Host"], params["Port"], idcode, remote_server
        )

    def _connect_xvc(
        self,
        xvc_host: str,
        xvc_port: str,
        idcode: int = 0,
        remote_server: str = "",
        done: request.DoneCallback = None,
    ):
        def done_connect_xvc(token, error, result):
            if log.is_domain_enabled("client", "INFO"):
                server = remote_server if remote_server else str(self)
                log.client.info(f"{server} connected to XVC Server {xvc_host}:{xvc_port}")
            done(token, error, result)

        if not remote_server:
            xvc_service = self.channel.getRemoteService("ContextXvc")
            ret = xvc_service.open(xvc_host, xvc_port, idcode, done_connect_xvc)
        else:
            cs_service = self.channel.getRemoteService("ChipScope")
            ret = cs_service.remote_connect_xvc(
                remote_server, xvc_host, xvc_port, idcode, done_connect_xvc
            )
        return ret

    def get_view(self, dm_view):
        dm_name = "chipscopy.dm." + dm_view if isinstance(dm_view, str) else dm_view.__name__
        if dm_name in self.views:
            return self.views[dm_name]

        if isinstance(dm_view, str):
            dm_view = __import__(dm_name, fromlist=[dm_view], globals=globals())
        if not protocol.isDispatchThread():
            manager = protocol.invokeAndWait(dm_view.add_manager_from_channel, self.channel)
        else:
            manager = dm_view.add_manager_from_channel(self.channel)
        view = ViewInfo(manager)
        self.views[dm_name] = view
        return view

    def target(self, target_id=None, cls=dm.Node, index=None, **kwargs):
        view = self.get_view(memory)

        return self._target(view, target_id, cls, index, **kwargs)

    def jtag_target(self, target_id=None, cls=dm.Node, index=None, **kwargs):
        view = self.get_view(jtag)

        return self._target(view, target_id, cls, index, **kwargs)

    def cs_target(self, target_id=None, cls=dm.Node, index=None, **kwargs):
        view = self.get_view("chipscope")

        return self._target(view, target_id, cls, index, **kwargs)

    def dc_target(self, target_id=None, cls=dm.Node, index=None, **kwargs):
        view = self.get_view("debugcore")

        return self._target(view, target_id, cls, index, **kwargs)

    jtarget = jtag_target

    def _target(self, view, target_id=None, cls=dm.Node, index=None, parent=None, **kwargs):
        if target_id is not None:
            ctx = ""
            if type(target_id) is int:
                ctx = view.target_ctx_map[target_id]
            else:
                ctx = target_id

            return view.get_node(ctx, cls)

        if kwargs or parent or cls != dm.Node:

            def match_func(node):
                for key, value in kwargs.items():
                    # if key == "parent" and isinstance(value, dm.Node):
                    #     if value.ctx != node.parent_ctx:
                    #         return False
                    if isinstance(value, str):
                        s = str(getattr(node, key))
                        if not re.search(value, s):
                            return False
                    elif getattr(node, key) != value:
                        return False
                return True

            if parent is not None:
                results = list(filter(match_func, view.get_children(parent)))
            else:
                results = list(filter(match_func, view.get_all()))
            if index is not None:
                return view.get_node(results[index].ctx, cls)
            else:
                nodes = []
                for res in results:
                    nodes.append(view.get_node(res.ctx, cls))
                if len(nodes) == 1:
                    nodes = nodes[0]
                return nodes

        if index is not None:
            parent = view
            return view[parent.children[index]]

        def print_targets(parent=None, level=1):
            for child in view.get_children(parent):
                name = child.Name if child.Name else child.ctx
                print("{}{} {}".format("\t" * level, child.target_id, name))
                print_targets(child, level + 1)

        print_targets()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
import time
from collections import defaultdict
from pathlib import Path
from typing import Optional, DefaultDict, Union, Dict, List

from chipscopy import CoreType
from chipscopy.api._detail.ltx import Ltx
from chipscopy.api._detail.trace import Trace
from chipscopy.api.containers import QueryList
from chipscopy.api.ddr.ddr import DDR
from chipscopy.api.ibert import IBERT
from chipscopy.api.ila import ILA
from chipscopy.api.memory import Memory
from chipscopy.api.noc.noc import NocPerfmon
from chipscopy.api.pcie import PCIe
from chipscopy.api.sysmon import Sysmon
from chipscopy.api.vio import VIO
from chipscopy.client import ServerInfo
from chipscopy.client.axis_ila_core_client import AxisIlaCoreClient
from chipscopy.client.axis_pcie_core_client import AxisPCIeCoreClient
from chipscopy.client.axis_trace_core_client import AxisTraceClient
from chipscopy.client.axis_vio_core_client import AxisVIOCoreClient
from chipscopy.client.core import CoreParent
from chipscopy.client.ddrmc_client import DDRMCClient
from chipscopy.client.ibert_core_client import IBERTCoreClient
from chipscopy.client.jtagdevice import JtagDevice
from chipscopy.client.noc_perfmon_core_client import NoCPerfMonCoreClient
from chipscopy.client.sysmon_core_client import SysMonCoreClient
from chipscopy.dm import chipscope, request, Node
from chipscopy.utils.printer import printer, PercentProgressBar
from chipscopy.api.device.device_identification import DeviceIdentification


class Device:
    """The Device class represents a single Xilinx Versal device.
    Device has methods for programming, memory access, and contains the detected
    collection of debug cores. The collection of detected devices is held in
    the session - see session.devices.
    """

    # This dict maps a core.type string to the correct node class type.
    CORE_TYPE_TO_CLASS_MAP = {
        "vio": AxisVIOCoreClient,
        "ila": AxisIlaCoreClient,
        "trace": AxisTraceClient,
        # todo: remove after hw_server is updated.
        "Core": AxisTraceClient,
        "pcie": AxisPCIeCoreClient,
        "npi_nir": NoCPerfMonCoreClient,
        "ibert": IBERTCoreClient,
        "sysmon": SysMonCoreClient,
        "ddrmc_main": DDRMCClient,
    }

    def __init__(
        self,
        *,
        hw_server: ServerInfo,
        cs_server: Optional[ServerInfo],
        device_identification: DeviceIdentification,
        disable_core_scan: bool,
    ):
        self.disable_core_scan = disable_core_scan
        self.cores_to_scan = {}  # set during discover_and_setup_cores

        self.hw_server = hw_server
        self.cs_server = cs_server
        self.device_identification = device_identification
        self.ltx = None

        self.debug_core_dict: DefaultDict[
            CoreType, QueryList[Union[VIO, ILA, PCIe, DDR, IBERT, NocPerfmon]]
        ] = defaultdict(QueryList)

        self.programming_done: bool = False
        self.programming_error: Optional[str] = None

        # The following filter_by becomes a dictionary with architecture, jtag_index, context, etc.
        # This is used when asking for a device by filter from the device list like:
        #    report_devices(session.devices.get(dna=".*1234"))
        # There is a problem with the new node identification as a list - need to flatten or something.
        # I don't think that the current filtering system will consider  list items in the dict.
        # TODO: address node_identification member of filter_by - maybe something other than loads below.
        self.filter_by = json.loads(repr(self.device_identification))

    def __str__(self):
        return str(self.device_identification)

    def __repr__(self):
        return repr(self.device_identification)

    def __getitem__(self, item):
        raw_json = repr(self.device_identification)
        parsed_json = json.loads(raw_json)
        return parsed_json[item]

    def _create_debugcore_wrapper(self, node):
        # Factory to build debug_core_wrappers
        # Instantiate debug core wrappers for a known core type. The wrapper gets attached to a
        # node for safe keeping. Any time we need to grab a debug core wrapper (like ILA)
        # it is available on the node using
        #     wrapper = self._get_client_wrapper(node)
        if node.type == "ila":
            debug_core_wrapper = ILA(node, self, ltx=self.ltx)
        elif node.type == "vio":
            debug_core_wrapper = VIO(node, ltx=self.ltx)
        elif node.type == "trace":
            debug_core_wrapper = Trace(node)
        # todo: remove after hw_server is updated.
        elif node.type == "Core":
            debug_core_wrapper = Trace(node)
        elif node.type == "pcie":
            debug_core_wrapper = PCIe(node)
        elif node.type == "npi_nir":
            debug_core_wrapper = NocPerfmon(node)
        elif node.type == "ibert":
            debug_core_wrapper = IBERT(node)
        elif node.type == "sysmon":
            debug_core_wrapper = Sysmon(node)
        elif node.type == "ddrmc_main":
            debug_core_wrapper = DDR(node)
        else:
            # TODO: extend this factory to other core types with a plugin model for additional types
            debug_core_wrapper = None
        return debug_core_wrapper

    def _find_all_debugcore_nodes(self) -> Dict:
        # Traverse the chipscope view for debug cores in this device.
        # Returns a dict mapping core_types to a list of found nodes.
        # Notes are returned in upgraded form.
        #   Example return:
        #         {
        #           AxisIlaCoreClient: [ila_node_1, ila_node_2, ...],
        #           AxisVIOCoreClient: [vio_node_1, vio_node_2, ...],
        #           ...
        #         }
        # The first time calling this function is slow because nodes have
        # not been upgraded yet. Slow is around 1 second. Subsequent calls are
        # faster, around 1 ms.
        #
        # Best to avoid calling this function in an inner loop.
        #
        cs_view = self.cs_server.get_view(chipscope)
        found_cores = defaultdict(list)
        # Using the self.chipscope_node restricts nodes to this device only
        for cs_node in cs_view.get_children(self.chipscope_node):
            if cs_node.type == "debug_hub":
                debug_hub_children = cs_view.get_children(cs_node)
                for child in debug_hub_children:
                    if self.cores_to_scan.get(child.type, True):
                        core_class_type = Device.CORE_TYPE_TO_CLASS_MAP.get(child.type, None)
                        if core_class_type and core_class_type.is_compatible(child):
                            upgraded_node = cs_view.get_node(child.ctx, core_class_type)
                            found_cores[core_class_type].append(upgraded_node)
            else:
                if self.cores_to_scan.get(cs_node.type, True):
                    core_class_type = Device.CORE_TYPE_TO_CLASS_MAP.get(cs_node.type, None)
                    if core_class_type and core_class_type.is_compatible(cs_node):
                        upgraded_node = cs_view.get_node(cs_node.ctx, core_class_type)
                        found_cores[core_class_type].append(upgraded_node)
        return found_cores

    def _get_debug_core_wrappers(self, core_type) -> QueryList:
        # Returns a QueryList of the debug_core_wrappers matching the given core_type
        #   core_type is a class from CORE_TYPE_TO_CLASS_MAP
        #   returned wrapper_type is ILA, VIO, etc wrapping the node
        #
        # debug_core_wrappers are initialized if they don't yet exist for the
        # node. After initialization, they are hung on the node for future reference.
        # This is convenient because their lifecycle is tied to the node.
        found_cores = QueryList()
        all_debug_core_nodes = self._find_all_debugcore_nodes()
        for node in all_debug_core_nodes.get(core_type, []):
            debug_core_wrapper = Device._get_client_wrapper(node)
            if debug_core_wrapper is None:
                # Lazy debugcore initialization here.
                # Initialization happens the first time we access the debug core.
                debug_core_wrapper = self._create_debugcore_wrapper(node)
            if debug_core_wrapper:
                self._set_client_wrapper(node, debug_core_wrapper)
            else:
                # TODO: Hasn't happened yet - is this an error, warning, or something else?
                pass
            # Now the node should have a proper debug_core_wrapper attached.
            if Device._get_client_wrapper(node):
                found_cores.append(Device._get_client_wrapper(node))
        return found_cores

    @property
    def jtag_node(self) -> JtagDevice:
        # Low level hw_server jtag node access
        if self.hw_server:
            discovery_node = self.device_identification.find_jtag_node(self.hw_server)
        else:
            raise RuntimeError("No hw_server server connection. Could not get jtag view")
        return discovery_node

    @property
    def debugcore_node(self) -> Node:
        # Low level hw_server debugcore access
        if self.hw_server:
            discovery_node = self.device_identification.find_debugcore_node(self.hw_server)
        else:
            raise RuntimeError("No hw_server server connection. Could not get debugcore view")
        return discovery_node

    @property
    def chipscope_node(self) -> CoreParent:
        # Low level cs_server chipscope access
        if self.cs_server:
            discovery_node = self.device_identification.find_chipscope_node(self.cs_server)
        else:
            raise RuntimeError("No chipscope server connection. Could not get chipscope view")
        return discovery_node

    @property
    def ibert_cores(self) -> QueryList[IBERT]:
        """Returns:
        list of detected IBERT cores in the device"""
        return self._get_debug_core_wrappers(IBERTCoreClient)

    @property
    def vio_cores(self) -> QueryList[VIO]:
        """Returns:
        list of detected VIO cores in the device"""
        return self._get_debug_core_wrappers(AxisVIOCoreClient)

    @property
    def ila_cores(self) -> QueryList[ILA]:
        """Returns:
        list of detected ILA cores in the device"""
        return self._get_debug_core_wrappers(AxisIlaCoreClient)

    @property
    def trace_cores(self) -> QueryList[Trace]:
        """Returns:
        list of detected Trace cores in the device"""
        return self._get_debug_core_wrappers(AxisTraceClient)

    @property
    def pcie_cores(self) -> QueryList[PCIe]:
        """Returns:
        list of detected PCIe cores in the device"""
        return self._get_debug_core_wrappers(AxisPCIeCoreClient)

    @property
    def noc_core(self) -> QueryList[NocPerfmon]:
        """Returns:
        List with one NOC core for the device"""
        return self._get_debug_core_wrappers(NoCPerfMonCoreClient)

    @property
    def ddrs(self):
        """Returns:
        list of detected DDRMC cores in the device"""
        return self._get_debug_core_wrappers(DDRMCClient)

    @property
    def sysmon_root(self) -> QueryList[Sysmon]:
        """Returns:
        List with one SysMon core for the device"""
        return self._get_debug_core_wrappers(SysMonCoreClient)

    @property
    # TODO: @chipscopy_deprecated("Use alternative session.mem.all() instead.")
    def memory_target_names(self):
        """Returns:
        list of supported memory targets"""
        valid_memory_targets = (
            r"(Versal .*)|(DPC)|(APU)|(RPU)|(PPU)|(PSM)|(Cortex.*)|(MicroBlaze.*)"
        )
        memory_target_names = list()
        memory_node_list = self.device_identification.get_memory_target_nodes(self.hw_server)
        for memory_node in memory_node_list:
            memory_target_name = memory_node.props.get("Name")
            if len(memory_target_name) > 0 and re.search(valid_memory_targets, memory_target_name):
                memory_target_names.append(memory_target_name)
        return memory_target_names

    @property
    def memory(self) -> QueryList[Memory]:
        """Returns:
        the memory access for the device"""
        target_nodes = self.device_identification.get_memory_target_nodes(self.hw_server)
        memory_node_list = QueryList()
        for node in target_nodes:
            if Memory.is_compatible(node):
                memory_node_list.append(node)
        return memory_node_list

    # ----------------------------------------------------------------------------------------------
    def reset(self):
        jtag_programming_node = self.device_identification.find_jtag_node(self.hw_server)
        assert jtag_programming_node is not None
        jtag_programming_node.future().reset()
        # Wait an additional 500ms just to be safe. I believe hw_server has a 200ms delay - had a couple of issues
        # so added some extra time.
        time.sleep(0.5)

    def program(
        self,
        pdi_file_path: Union[str, Path],
        *,
        skip_reset: bool = False,
        show_progress_bar: bool = True,
        done: request.DoneFutureCallback = None,
    ):
        """
        Program the device with a given PDI file.

        Args:
            pdi_file_path: PDI file path
            skip_reset: False = Do not reset device prior to program
            show_progress_bar: False if the progress bar doesn't need to be shown
            done: Optional async future callback
        """
        program_future = request.CsFutureSync(done)

        if isinstance(pdi_file_path, str):
            pdi_file_path = Path(pdi_file_path)

        if not pdi_file_path.exists():
            raise FileNotFoundError(f"PDI file at {str(pdi_file_path.resolve())} doesn't exists!")

        printer(f"Programming PDI file {str(pdi_file_path.resolve())}\n", level="info")

        progress = PercentProgressBar()
        progress.add_task(
            description="Device program progress",
            status=PercentProgressBar.Status.STARTING,
            visible=show_progress_bar,
        )

        def progress_update(future):
            progress.update(
                completed=future.progress * 100, status=PercentProgressBar.Status.IN_PROGRESS
            )

        def done_programming(future):
            if future.error is not None:
                progress.update(status=PercentProgressBar.Status.ABORTED)
            else:
                progress.update(completed=100, status=PercentProgressBar.Status.DONE)

            self.programming_error = future.error
            self.programming_done = True
            if future.error:
                program_future.set_exception(future.error)
            else:
                program_future.set_result(None)

        def finalize_program():
            if self.programming_error is not None:
                raise RuntimeError(self.programming_error)

        # Go through list of possible jtag configurable devices - when we find our matching context,
        # program it. This is versal specific. It will need work when we add ultrascale.
        # One problem at a time...
        jtag_programming_node = self.device_identification.find_jtag_node(self.hw_server)
        assert jtag_programming_node is not None
        # TODO: Log the jtag node information to a log or something that we are about to program
        if not skip_reset:
            jtag_programming_node.future().reset()

        jtag_programming_node.future(
            progress=progress_update, done=done_programming, final=finalize_program
        ).config(str(pdi_file_path.resolve()))

        return program_future if done else program_future.result

    @staticmethod
    def _set_client_wrapper(node, api_client_wrapper):
        # Tacks a wrapper onto a node for tracking. Always call this
        # instead of setting directly on node - in case we change the
        # client wrapper tracking process.
        node.api_client_wrapper = api_client_wrapper

    @staticmethod
    def _get_client_wrapper(node):
        # Gets the tacked on wrapper from a node.
        return node.api_client_wrapper

    def discover_and_setup_cores(
        self,
        *,
        hub_address_list: Optional[List[int]] = None,
        ltx_file: str = None,
        ila_scan: bool = True,
        vio_scan: bool = True,
        noc_scan: bool = False,
        ddr_scan: bool = True,
        pcie_scan: bool = True,
        ibert_scan: bool = False,
        sysmon_scan: bool = False,
    ):
        """Scan device for debug cores. This may take some time depending on
        what gets scanned.

        Args:

            hub_address_list: List of debug hub addresses to scan
            ltx_file: LTX file (which contains debug hub addresses)
            ila_scan: True=Scan Device for ILAs
            vio_scan: True=Scan Device for VIOs
            ibert_scan: True=Scan Device for IBERTs
            pcie_scan: True=Scan Device for PCIEs
            noc_scan: True=Scan Device for NOC
            ddr_scan: True=Scan Device for DDRs
            sysmon_scan: True=Scan Device for System Monitor
        """

        # TODO: Cleanup: Could we pass the <core>_scan optional args as a dict or something?

        # Selectively disable scanning of cores depending on what comes in
        # This is second priority to the disable_core_scan in __init__.
        self.cores_to_scan = {}
        if self.disable_core_scan:
            for core_name in Device.CORE_TYPE_TO_CLASS_MAP.keys():
                self.cores_to_scan[core_name] = False
        else:
            self.cores_to_scan = {
                "ila": ila_scan,
                "vio": vio_scan,
                "npi_nir": noc_scan,
                "ddrmc_main": ddr_scan,
                "pcie": pcie_scan,
                "ibert": ibert_scan,
                "sysmon": sysmon_scan,
            }

        if not self.cs_server:
            raise RuntimeError("cs_server is not connected")

        self.ltx = None
        # Erase any old cores lingering from a previously detected run
        self.debug_core_dict.clear()

        # Make sure debug_hub_addrs is a list of hub addresses.
        # It may come as an int or a list or None
        if isinstance(hub_address_list, int):
            hub_address_list = [hub_address_list]
        if not hub_address_list:
            hub_address_list = []

        if ltx_file:
            self.ltx = Ltx()
            self.ltx.parse_file(ltx_file)
            hub_address_list.extend(self.ltx.get_debug_hub_addresses())
            # Remove any duplicate hub addresses in list, in case same one
            # was passed to function as well as in ltx
            # Below is a quick python trick to uniquify a list.
            hub_address_list = list(dict.fromkeys(hub_address_list))

        # Set up all debug cores in hw_server for the given hub addresses
        # Do this by default unless the user has explicitly told us not to with
        # "disable_core_scan" as an argument.
        if not self.disable_core_scan:
            # setup_debug_cores can take a while...
            self.chipscope_node.setup_cores(debug_hub_addrs=hub_address_list)

    def memory_write(self, address: int, values: List[int], *, size="w", target=None):
        """Write list of values to given memory address. This method is slow because it locates the memory context
        in the target tree every time. If you want to execute a large number of memory operations, grab a memory
        context and do the operations in batch.

        Note: This method should not be used in inner loops. It is not the fastest because it looks up
              the memory target every time. Inner loops should just call the find_memory_target once.
        """
        node = self.device_identification.find_memory_target(self.hw_server, target)
        node.memory_write(address=address, values=values, size=size)

    def memory_read(self, address: int, num: int = 1, *, size="w", target=None) -> List[int]:
        """Read num values from given memory address.  This method is slow because it locates the memory context
        in the target tree every time. If you want to execute a large number of memory operations, grab a memory
        context and do all the operations in batch.

        Note: This method should not be used in inner loops. It is not the fastest because it looks up
              the memory target every time. Inner loops should just call the find_memory_target once.
        """
        node = self.device_identification.find_memory_target(self.hw_server, target)
        retval = node.memory_read(address=address, num=num, size=size)
        return retval

    def _dpc_read_bytes(
        self,
        address: int,
        data: bytearray,
        *,
        offset: int = 0,
        byte_size: int = None,
        show_progress_bar: bool = True,
    ):
        """
        Reads byte data from memory, using the DPC.

        Args:
            address (int): Start address
            data (bytearray): result buffer which gets filled with read data.
            offset (int): start index in *data* to start write to.
            byte_size (int): Number of bytes to read. Default is the size of the *data* buffer.
            show_progress_bar (bool): Show if True.
        """
        node = self.debugcore_node
        if not show_progress_bar:
            node.read_bytes(address, data, offset, byte_size)
            return

        # With progress bar
        transfer_byte_count = byte_size if byte_size else len(data)
        desc = (
            f"Transferring {transfer_byte_count/(1024 * 1024):0.1f} MB "
            f"[bold](Read {hex(address)})[/]"
        )

        bar = PercentProgressBar()
        bar.add_task(description=desc, status=PercentProgressBar.Status.STARTING)

        try:
            node.read_bytes(
                address,
                data,
                offset,
                byte_size,
                progress=lambda progress: bar.update(
                    completed=int(progress * 100), status=PercentProgressBar.Status.IN_PROGRESS
                ),
            )
        except Exception as ex:
            bar.update(status=PercentProgressBar.Status.ABORTED)
            raise ex

        bar.update(status=PercentProgressBar.Status.DONE)

    def _dpc_write_bytes(
        self,
        address: int,
        data: bytearray,
        *,
        offset: int = 0,
        byte_size: int = None,
        show_progress_bar: bool = True,
    ):
        """
        Write byte data from memory, using the DPC.

        Args:
            address (int): Start address
            data (bytearray): data to write.
            offset (int): start index in *data* to start write from.
            byte_size (int): Number of bytes to write. Default is the size of the *data* buffer.
            show_progress_bar (bool): Show if True.
        """
        node = self.debugcore_node
        if not show_progress_bar:
            node.write_bytes(address, data, offset, byte_size)
            return

        # With progress bar
        transfer_byte_count = byte_size if byte_size else len(data)
        desc = (
            f"Transferring {transfer_byte_count/(1024 * 1024):0.1f} MB "
            f"[bold](Write {hex(address)})[/]"
        )

        bar = PercentProgressBar()
        bar.add_task(description=desc, status=PercentProgressBar.Status.STARTING)

        try:
            node.write_bytes(
                address,
                data,
                offset,
                byte_size,
                progress=lambda progress: bar.update(
                    completed=int(progress * 100), status=PercentProgressBar.Status.IN_PROGRESS
                ),
            )
        except Exception as ex:
            bar.update(status=PercentProgressBar.Status.ABORTED)
            raise ex

        bar.update(status=PercentProgressBar.Status.DONE)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. |Node|      replace:: :class:`Node`
.. |CsManager| replace:: :class:`CsManager`

Node
^^^^^^^^^^^^^^^
.. autoclass:: Node
    :members:

CsManager
^^^^^^^^^^^^^^^
.. autoclass:: CsManager
    :members:

NodeListener
^^^^^^^^^^^^
.. autoclass:: NodeListener
    :members:

CsManagerListener
^^^^^^^^^^^^^^^^^
.. autoclass:: CsManagerListener
    :members:

"""

import re
from typing import Dict, Type, Any, List, Set, Tuple, NewType, Callable
from collections import deque
from chipscopy.tcf import protocol

NodePropsCallback = NewType("NodePropsCallback", Callable[["Node", Set[str]], None])

_manager_listeners = []

_managers = {}

_manager_match = re.compile(r"\[(.+)\]-(.+)")


def get_manager(name: str) -> "CsManager":
    """
    Gets manager from name (usually based on Channel ID)
    :param name: Name of manager
    :return: Manager object
    """
    return _managers.get(name)


def add_manager(manager: "CsManager"):
    """
    Adds CsManager to map of managers
    :param manager: Manager to add
    """
    _managers[manager.name] = manager


def remove_manager(manager: "CsManager"):
    """
    Removes CsManager from map of managers
    :param manager: Manager to remove
    """
    manager.invalidate()
    del _managers[manager.name]


def node_id_from_context(node: "Node"):
    """
    Creates a ChipScope Server unique ID from the manager's id and node context handle.
    :param node: Node from which to create unique ID
    """
    return "[{}]-{}".format(node.manager.name if node.manager else "ID", node.ctx)


def parse_node_id(node_id: str) -> Tuple[str, str]:
    """
    Parses a node_id and retrieves the manager's name and the node's context handle
    :param node_id: Node ID
    :return: Manager name and node context handle
    """
    # mn, _, ctx = node_id.partition("-")
    m = _manager_match.match(node_id)
    mn = ""
    ctx = ""
    if m:
        mn, ctx = m.group(1, 2)
    return mn, ctx


class Node(object):
    """
    A node is a generic hardware Context.  Node base class provides storage for Context properties provided by the
    hw_server and ChipScope services.  Node objects are intended to be 'upgraded' to Node subclasses to add
    functionality for a particular Context type.

    .. attribute:: node.key

        Returns attribute normal if exists.  Else returns property if exists, or None if not.  Can not be used for
        setting properties.

    .. method:: node[key]

        Return property of node from key.  Raises |KeyError| if property does not exist.

    .. method:: node[key] = value

        Sets value of property for given key.  It also auto updates listeners.
    """

    def __init__(self, ctx: str = "", parent_ctx: str = "", manager: "CsManager" = None):
        self.ctx = ctx
        self.parent_ctx = parent_ctx
        self.manager = manager
        self.children = []
        self._props = {}
        self._updated_props = set()
        self.dirty = False
        self.pending = set()
        self.invalid = False
        self.listeners: NodePropsCallback = list()

    def post_init(self):
        """
        Used to handle initialization where __init__ was called without any parameters were set
        """
        pass

    def de_init(self):
        """
        Used to clean up when this class is being switched away from
        :return:
        """
        pass

    @property
    def props(self) -> Dict[str, Any]:
        """
        Accessor for full list of properties
        :return: Properties
        """
        return self._props

    @property
    def size(self) -> int:
        """
        :return: The number of properties
        """
        return len(self._props)

    @property
    def is_valid(self) -> bool:
        """
        A Node is valid while still existing in the hw_server.  Once the node is removed in the hw_server the
        node here is invalidated rather than removed.  The node is removed once a replacement is found in the hw_server.
        """
        return not self.invalid

    @staticmethod
    def is_compatible(node: "Node") -> bool:
        """
        Overridden by subclasses to test if this node can be converted to subclass instance.  Usually tests available
        properties of the node and/or position in the tree.
        :param node: node instance to test
        :return: True if compatible
        """
        return True

    @property
    def queue_group(self) -> Any:
        """
        Queue groups are used to control the flow of requests on one or more nodes.  Requests on the same queue
        group are handled synchronously to maintain consistent state.  By default the node is in its own queue group to
        maximize the number of asynchronous operations permitted across all nodes.

        :return: Context handle of parent of the queue group
        """
        return self.ctx

    def update(self, props: Dict[str, Any]):
        """
        Updates properties of node and notifies any listeners of the node's changed properties.

        :param props: Properties to update.
        """
        self._props.update(props)
        keys = set(props.keys())
        if self.manager and keys not in self._updated_props:
            self._updated_props.update(keys)
            self._props_updated()

    def __getattr__(self, attr):
        # NOTE uncomment these lines if having trouble debugging __init__() in child class
        if "_props" not in self.__dict__:
            return None
        if attr in self._props:
            return self._props.get(attr)

    def __getitem__(self, item):
        return self._props[item]

    def __setitem__(self, key, value):
        self._props[key] = value
        if self.manager and key not in self._updated_props:
            self._updated_props.add(key)
            self._props_updated()

    def __str__(self):
        return "{}({}, {} props)".format(self.__class__.__name__, self.ctx, len(self._props))

    @property
    def is_ready(self):
        """
        Checks to see if this node is currently available for operation.  Generally will not be ready if there are
        active operations being run on this node (e.g. out standing TCF command to the hw_server)
        :return: True if ready and available
        """
        return not self.pending

    def add_pending(self, token):
        self.pending.add(token)
        return token

    def remove_pending(self, token):
        try:
            self.pending.remove(token)
        except KeyError:
            pass

    def invalidate(self, notify=True):
        if not self.invalid:
            self.invalid = True
            if self.manager:
                for child in self.manager.get_children(self):
                    child.invalidate()
                if notify:
                    self.manager.notify_node_removed(self)
            self.listeners = set()

    def _props_updated(self):
        if not self.dirty:
            self.dirty = True
            if self.manager:
                self.manager.notify_node_changed(self)

    def clear_update(self):
        # clears parameters used to notify on updates
        self.dirty = False
        self._updated_props.clear()

    def add_listener(self, listener: NodePropsCallback):
        self.listeners.append(listener)

    def remove_listener(self, listener: NodePropsCallback):
        self.listeners.remove(listener)

    def call_listeners(self):
        for listener in self.listeners:
            listener(self, self._updated_props)


class NodeAutoUpgrader(object):
    """ Used by a Manager of a view to auto upgrade a node to specific classes """

    def __call__(self, node: Node, **kwargs):
        """ Called by view manager """
        return node


class CsManager(Node):
    """
    Manages Context |Node| objects in a directly accessible dict by their ctx handles and provides means to navigate by
    a node's parent or children.

    .. method:: manager[ctx]

        Return node from given ctx.  Raises |KeyError| if node does not exist.
    """

    node_auto_upgrader: NodeAutoUpgrader = None

    def __init__(self, name: str = "", channel=None):
        super(CsManager, self).__init__(ctx="", manager=self)
        self.channel = channel
        self._nodes = {"": self}  # the manager is always the default "" parent
        self._node_listeners = []
        self._added_nodes = deque()
        self._removed_nodes = deque()
        self.name = name
        self.default_node_cls = Node
        for listener in _manager_listeners:
            listener.manager_added(self)

    def add_node(self, ctx: str, parent_ctx: str = "") -> Node:
        """
        Creates and Adds new node to manager from given ctx and parent's ctx.

        :param ctx: Context handle of new node
        :param parent_ctx: Context handle of parent
        :return: node added
        """
        props = None
        node = self._nodes.get(ctx)
        if node and node.parent_ctx != parent_ctx:
            props = node.props
            node.invalidate()
            node = None

        if node:
            if node.invalid:
                node.invalid = False
                self.notify_node_added(node)
            return node

        # create and add to manager
        node = self.default_node_cls(ctx, parent_ctx, self)
        if props:
            node.props.update(props)
        self._nodes[ctx] = node

        # add to parent's children if not already
        parent = self._nodes.get(parent_ctx)
        if parent:
            if ctx not in parent.children:
                parent.children.append(ctx)

        node.post_init()

        # notify node added
        self.notify_node_added(node)

        return node

    def remove_node(self, ctx: str):
        """
        Removes node and its children from manager at given ctx and removes the node from its parent.

        :param ctx: Context handle of node
        """
        try:
            node = self._nodes.pop(ctx)
        except KeyError:
            return

        node.invalid = True

        # remove from parent
        parent = self._nodes.get(node.parent_ctx)
        if parent:
            try:
                parent.children.remove(ctx)
            except ValueError:  # pragma: no cover
                pass

        self.notify_node_removed(node)

        # remove all children
        for child in node.children:
            self.remove_node(child)

    def remove_all(self):
        """
        Removes all nodes from manager
        """
        while self.children:
            self.remove_node(self.children[0])

    def upgrade_node(self, node: Node, upgrade_cls: type, **kwargs) -> Node:
        """
        Converts node to a Node subclass and maintains node properties and position in manager.

        :param node: Node to convert
        :param upgrade_cls: Node subclass
        :return: new node instance
        """
        node.de_init()
        new = upgrade_cls(**kwargs)
        new.__dict__.update(node.__dict__)
        self._nodes[new.ctx] = new
        new.post_init()
        node.invalid = True
        return new

    def get_children(self, parent: str or Node = "") -> Node:
        """
        Iterates through list of children of given parent

        :param parent: Context handle or node object of parent
        :yield: child node
        """
        if parent is None:
            parent = self
        if isinstance(parent, str):
            parent = self[parent]
        if not parent.children:
            return None

        for child in parent.children:
            yield self._nodes.get(child)

    def get_all(self) -> Node:
        """
        Iterates through all nodes in manager.

        :yield: node object
        """
        for node in self._nodes.values():
            yield node

    def get_node(self, ctx: str, cls: Type[Node] = Node) -> Node or None:
        """
        Gets the node instance of a given class if compatible.  If the node is not currently of the class it's
        'upgraded' in the CsManager to be an instance of that class.  The property values are maintained.

        :param ctx: Context handle of node
        :param cls: Desired class of node
        :return: node of class if compatible; None if not
        """
        node = self._nodes.get(ctx)

        if not node:
            return None

        if isinstance(node, cls):
            return node

        if cls.is_compatible(node):
            return self.upgrade_node(node, cls)

        return None

    def set_children(self, parent: Node or str, children: List[str]):
        """
        Sets the children list of a parent node.  If existing child is not on the new list of children it is removed.

        :param parent: Parent context node
        :param children: List of context handles of new children
        """
        if isinstance(parent, str):
            parent = self[parent]

        remove_children = [child for child in parent.children if child not in children]
        for child in remove_children:
            self.remove_node(child)

        parent.children = children

    def notify_node_changed(self, node: Node):
        if self._node_listeners:
            protocol.invokeLater(self._node_updated, node.ctx)
        else:
            node.clear_update()

    def _node_updated(self, node_ctx: str):
        node = self._nodes.get(node_ctx)
        if node:
            if node._updated_props:
                for listener in self._node_listeners:
                    node.call_listeners()
                    listener.node_changed(node, node._updated_props)
            node.clear_update()

    def notify_node_added(self, node: Node):
        if self._node_listeners:
            if node.ctx not in self._added_nodes:
                should_post = len(self._added_nodes) == 0
                self._added_nodes.append(node.ctx)
                if should_post:
                    protocol.invokeLater(self._node_added)
        else:
            node.clear_update()

    def _node_added(self):
        nodes = []
        if self._added_nodes:
            for ctx in self._added_nodes:
                node = self._nodes.get(ctx)
                if node:
                    nodes.append(node)
                    node.clear_update()
        if nodes:
            for listener in self._node_listeners:
                listener.nodes_added(nodes)
        self._added_nodes.clear()

    def notify_node_removed(self, node: Node):
        if self._node_listeners:
            if node not in self._removed_nodes:
                should_post = len(self._removed_nodes) == 0
                self._removed_nodes.append(node)
                if should_post:
                    protocol.invokeLater(self._node_removed)

    def _node_removed(self):
        if self._removed_nodes:
            for listener in self._node_listeners:
                listener.nodes_removed(self._removed_nodes)
        self._removed_nodes.clear()

    @property
    def size(self) -> int:
        """
        :return: The number of nodes.
        """
        return len(self._nodes)

    def __getitem__(self, item) -> Node:
        return self._nodes[item]

    def __str__(self):
        return "{}({}, {} nodes)".format(self.__class__.__name__, self.name, len(self._nodes))

    @property
    def queue_group(self):
        return self

    def invalidate(self):
        self.invalid = True
        for listener in _manager_listeners:
            listener.manager_removed(self)

        for child in self.get_children():
            child.invalidate()

    def add_node_listener(self, listener: "NodeListener"):
        """
        Adds |Node| event listener to manager.
        :param listener: Node listener to add
        """
        if listener not in self._node_listeners:
            self._node_listeners.append(listener)

    def remove_node_listener(self, listener: "NodeListener"):
        """
        Removes |Node| event listener from manager.
        :param listener: Node listener to remove
        """
        self._node_listeners.remove(listener)


class NodeListener(object):
    """
    |Node| event listener catching when nodes are created, removed, or changed
    """

    def nodes_added(self, nodes: List[Node]):
        """
        Called when a node was added to the manager.

        :param nodes: List of context nodes added
        """
        pass

    def node_changed(self, node: Node, updated_keys: Set[str]):
        """
        Called when a property was changed in a node.
        :param node: Context node changed
        :param updated_keys: List of property keys that have changed
        """
        pass

    def nodes_removed(self, nodes: List[Node]):
        """
        Called when a node is invalidated.
        :param nodes: List of context nodes removed
        """
        pass


class CsManagerListener(object):
    """
    |CsManager| event listener catching when CsManagers are created or removed
    """

    def manager_added(self, cs_manager: CsManager):
        """
        Called when CsManager added.

        :param cs_manager: A manager that has been created
        """
        pass

    def manager_removed(self, cs_manager: CsManager):
        """
        Called when CsManager removed.

        :param cs_manager: A manager that has been removed
        """
        pass

    @staticmethod
    def add_listner(listener: "CsManagerListener"):
        """
        Adds CsManagerListener to list
        :param listener: Listener to add
        """
        if listener not in _manager_listeners:
            _manager_listeners.append(listener)

    @staticmethod
    def remove_listener(listener: "CsManagerListener"):
        """
        Removes CsManagerListener from list of listeners
        :param listener: Listener to remove
        """
        _manager_listeners.remove(listener)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from chipscopy.tcf import protocol
from chipscopy.tcf.channel import ChannelListener
from chipscopy.proxies.JtagProxy import JtagProxy as JtagService
from chipscopy.utils.logger import log
from . import CsManager, Node, _managers, add_manager, remove_manager

MANAGER_TYPE = "jtag"

csm_pattern = re.compile(r".+\." + MANAGER_TYPE)


def get_managers():
    for name, csm in _managers.items():
        if csm_pattern.match(name):
            yield csm


def add_manager_from_channel(channel):
    jtag = channel.getRemoteService(JtagService)
    if not jtag:
        raise Exception("Server does not have Jtag Service")

    # set up memory view manager
    name = "{}.{}".format(channel.remote_peer.getID(), MANAGER_TYPE)
    for manager in get_managers():
        if manager.name == name:
            if manager.channel != channel:
                remove_manager(manager)
                break
            return manager

    cs_manager = JtagManager(name, channel)
    add_manager(cs_manager)

    # add channel listener for disconnection event
    channel.addChannelListener(ManagerChannelListener(cs_manager))

    # trigger update
    cs_manager.update_nodes()
    return cs_manager


class JtagListener(object):
    def __init__(self, manager):
        self.manager = manager

    def contextAdded(self, contexts):
        log.jtag2.debug(f"Jtag Event contextAdded {[c.ID for c in contexts]}")
        for context in contexts:
            parent_ctx = context.ParentID
            if parent_ctx is None:
                parent_ctx = ""
            node = self.manager.add_node(context.ID, parent_ctx)
            node.jtag = context
            node.update(context.props)
            if self.manager.node_auto_upgrader:
                self.manager.node_auto_upgrader(node)

    def contextChanged(self, contexts):
        log.jtag2.debug(f"Jtag Event contextChanged {[c.ID for c in contexts]}")
        for context in contexts:
            parent_ctx = context.ParentID
            if parent_ctx is None:
                parent_ctx = ""
            node = self.manager.add_node(context.ID, parent_ctx)
            node.jtag = context
            node.update(context.props)

    def contextRemoved(self, context_ids):
        log.jtag2.debug(f"Jtag Event contextRemoved {context_ids}")
        for context_id in context_ids:
            try:
                node = self.manager[context_id]
                node.invalidate()
            except KeyError:
                pass


class JtagManager(CsManager):
    def __init__(self, name: str = "", channel=None):
        super(JtagManager, self).__init__(name, channel)
        jtag = self.channel.getRemoteService(JtagService)
        jtag.add_listener(JtagListener(self))

    def update_nodes(self):
        self._update_children(self)

    def _update_children(self, parent):
        jtag = self.channel.getRemoteService(JtagService)
        cs_manager = self

        def done_get_children(token, error, context_ids):
            if not error:
                cs_manager.set_children(parent, context_ids)

                for context in context_ids:
                    node = cs_manager.add_node(context, parent.ctx)
                    cs_manager._update_static_info(node)
                    cs_manager._update_children(node)
            cs_manager.remove_pending(token)

        if jtag:
            self.add_pending(jtag.get_children(parent.ctx, done_get_children))

    def _update_static_info(self, node: Node):
        jtag = self.channel.getRemoteService(JtagService)

        def done_get_context(token, error, context):
            if not error:
                props = context.props
                node.update(props)
                if self.node_auto_upgrader:
                    self.node_auto_upgrader(node)
            node.remove_pending(token)

        if jtag:
            node.add_pending(jtag.get_context(node.ctx, done_get_context))


class ChannelOpenListener(protocol.ChannelOpenListener):
    def onChannelOpen(self, channel):
        add_manager_from_channel(channel)


class ManagerChannelListener(ChannelListener):
    def __init__(self, cs_manager: JtagManager):
        self.cs_manager = cs_manager

    def onChannelClosed(self, error):
        if self.cs_manager:
            self.cs_manager.invalidate()
        self.cs_manager.channel.removeChannelListener(self)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

from chipscopy.api.ibert.aliases import (
    RX_BER,
    RX_LINE_RATE,
    RX_PATTERN_CHECKER_ERROR_COUNT,
    RX_RECEIVED_BIT_COUNT,
    RX_STATUS,
)
from chipscopy.api.ibert.link.group import LinkGroup
from chipscopy.api.ibert.rx import RX
from chipscopy.api.ibert.tx import TX
from chipscopy.utils.printer import printer
from rich.box import SQUARE as BOX_SQUARE
from rich.table import Table

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.eye_scan import EyeScan


@dataclass
class Link:
    """
    Class for interacting with links.
    **Please do not** create an instance of this class directly. Please use the factory method
    :py:func:`~chipscopy.api.ibert.create_links` instead.
    """

    rx: Optional[RX]
    """RX attached to this link"""

    tx: Optional[TX]
    """TX attached to this link"""

    name: str
    """Name of this link"""

    handle: str = field(default="")
    """
        Handle in the format ``tx.handle-->rx.handle``. 
        If TX/RX in unknown, replace ``tx/rx.handle`` with ``UnknownTX/RX``
    """

    filter_by: Dict[str, Any] = field(default_factory=dict)

    link_group: Optional[LinkGroup] = None
    """Link group this link belongs to"""

    # NOTE - Either TX or RX can be unknown, not both!
    #  It is assumed the link manager will perform this check before creating link.

    def __repr__(self) -> str:
        return self.name

    def __post_init__(self):
        # Both RX and TX can't be unknown
        if self.rx is None and self.tx is None:
            raise ValueError("Both RX and TX cant be unknown!")

        if self.tx is not None:
            if self.tx.link is not None:
                raise ValueError(f"{self.tx.handle} is already part of {self.tx.link.name}!")

            self.tx.link = self
            self.handle = self.tx.handle
        else:
            self.handle = "UnknownTX"

        self.handle += "-->"

        if self.rx is not None:
            if self.rx.link is not None:
                raise ValueError(f"{self.rx.handle} is already part of {self.rx.link.name}!")

            self.rx.link = self
            self.handle += self.rx.handle
        else:
            self.handle = "UnknownRX"

        # This is used by the filter_by method in QueryList
        self.filter_by = {
            "rx": self.rx,
            "tx": self.tx,
            "name": self.name,
            "link_group": self.link_group,
        }

    def invalidate(self):
        if self.rx is not None:
            self.rx.link = None
        if self.tx is not None:
            self.tx.link = None
        if self.link_group is not None:
            self.link_group.remove(self)

        self.rx = None
        self.tx = None
        self.name = ""
        self.handle = ""
        self.link_group = None

    @property
    def eye_scan(self) -> Optional["EyeScan"]:
        """
        Get the :py:class:`~chipscopy.api.ibert.eye_scan.EyeScan` object attached to
        the :py:class:`~RX` that is part of the link.

        Returns:
            Eye scan object if available in RX

        """
        return self.rx.eye_scan if self.rx is not None else None

    @property
    def ber(self) -> float:
        """
        Refresh and get the BER value.

        .. note::
            There might be some loss of precision when converting the BER value from str to float.

        Returns:
            BER str value converted to float

        """
        _, prop_value = self.rx.property.refresh(self.rx.property_for_alias[RX_BER]).popitem()
        return float(prop_value)

    @property
    def status(self) -> str:
        """
        Refresh and get the status.

        Returns:
            Link status
        """
        _, prop_value = self.rx.property.refresh(self.rx.property_for_alias[RX_STATUS]).popitem()
        return prop_value

    @property
    def line_rate(self) -> str:
        """
        Refresh and get the line rate of the RX

        Returns:
            Line rate
        """
        _, prop_value = self.rx.property.refresh(self.rx.property_for_alias[RX_LINE_RATE]).popitem()
        return prop_value

    @property
    def bit_count(self):
        """
        Refresh and get the bits received by the RX

        Returns:
            Bit count
        """
        _, prop_value = self.rx.property.refresh(
            self.rx.property_for_alias[RX_RECEIVED_BIT_COUNT]
        ).popitem()
        return prop_value

    @property
    def error_count(self):
        """
        Refresh and get the error counter value

        Returns:
            Error counter value
        """
        _, prop_value = self.rx.property.refresh(
            self.rx.property_for_alias[RX_PATTERN_CHECKER_ERROR_COUNT]
        ).popitem()
        return prop_value

    def generate_report(self):
        """
        Generate a report for this link and send it to the "printer" for printing
        """
        report = Table(title=f'Link "{self.handle}" report', box=BOX_SQUARE)
        report.add_column("Property", justify="right")
        report.add_column("Value", justify="left")

        report.add_row(f"Name", f"{self.name}")
        report.add_row(f"RX", f"{None if self.rx is None else self.rx.handle}")
        report.add_row(f"TX", f"{None if self.tx is None else self.tx.handle}")
        report.add_row(
            f"Link Group", f"{None if self.link_group is None else self.link_group.name}"
        )

        report.add_row("", "")

        if self.rx is not None:
            values = self.rx.property.get(list(self.rx.property_for_alias.values()))
            for alias, prop in self.rx.property_for_alias.items():
                report.add_row(f"RX {alias}", f"{values[prop]}")

        report.add_row("", "")

        if self.tx is not None:
            values = self.tx.property.get(list(self.tx.property_for_alias.values()))
            for alias, prop in self.tx.property_for_alias.items():
                report.add_row(f"TX {alias}", f"{values[prop]}")

        printer("\n")
        printer(report)
        printer("\n")

    def refresh(self):
        """
        Refresh the attributes
        """
        if self.rx is not None:
            self.rx.property.refresh(list(self.rx.property_for_alias.values()))
        if self.tx is not None:
            self.tx.property.refresh(list(self.tx.property_for_alias.values()))



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import namedtuple
from pathlib import Path

import importlib_metadata
import os
import errno
import inspect

# NOTE - Only utility functions that are user visible should be added here to avoid circular imports
from chipscopy.api import CoreType
from chipscopy.api.session import create_session, delete_session
from chipscopy.dm.request import null_callback
from chipscopy.api.report import report_versions, report_devices, report_hierarchy

__author__ = "Xilinx, Inc."
__copyright__ = "Copyright 2020, Xilinx, Inc."
__email__ = "labtools@xilinx.com"
try:
    __version__ = importlib_metadata.version(__package__)
except importlib_metadata.PackageNotFoundError:  # for frozen app support, enter correct version here
    __version__ = "XXXXXXX"


def get_examples_dir_or_die():
    # Examples are provided for Jupyter and Commandline flows. This returns the
    # examples directory in both flows.

    # First priority - locally installed examples with the default directory name
    # Case for most users that installed examples.
    examples_dir = os.path.realpath(os.getcwd() + "./chipscopy-examples")
    if os.path.isdir(examples_dir):
        return examples_dir

    # If we get here, try to find examples in the site-libraries area or wherever
    # we are running chipscopy from. This makes life easier for developers working
    # who want to use examples directly from the chipscopy/examples directory.
    if os.path.isfile(inspect.getfile(inspect.currentframe())):
        # __file__ does not work in jupyter flows
        examples_dir = os.path.realpath(
            os.path.dirname(inspect.getfile(inspect.currentframe())) + "/examples"
        )
    else:
        # Likely in a jupyter environment - examples will likely be in the same dir as the cwd
        examples_dir = os.path.dirname(os.path.realpath(os.getcwd() + "./examples"))
    if not os.path.isdir(examples_dir):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), examples_dir)
    return examples_dir


def _get_design_files_in_dir(p: Path) -> namedtuple:
    ProgrammingFiles = namedtuple("ProgrammingFiles", "programming_file probes_file hwh_file")
    hwh_file = None
    ltx_file = None
    programming_file = None
    if p.is_dir():
        hwh_files = list(p.glob("*.hwh"))
        ltx_files = list(p.glob("*.ltx"))
        pdi_files = list(p.glob("*.pdi"))
        bit_files = list(p.glob("*.bit"))

        if hwh_files:
            hwh_file = str(hwh_files[0])
        if ltx_files:
            ltx_file = str(ltx_files[0])
        if pdi_files:
            programming_file = str(pdi_files[0])
        elif bit_files:
            programming_file = str(bit_files[0])
    return ProgrammingFiles(programming_file, ltx_file, hwh_file)


def get_design_files(board_path: str) -> namedtuple:
    # board_path like "vck190/production/2.0/ks_demo" pointing to a valid board design
    # in one of the repository areas. Must have ltx and bit or pdi files.
    # Prioritized look for designs
    # 1. In current directory deployed by chipscopy-get-examples
    # 2. site-libraries or chipscopy dev area
    # 3. jupyter notebook cwd
    # 4. internal-examples area

    design_dirs = []
    examples_dir = None
    try:
        examples_dir = get_examples_dir_or_die()
    except FileNotFoundError:
        pass

    if examples_dir:
        p = Path(examples_dir)
        p1 = p / "designs" / board_path
        if p1.is_dir():
            design_dirs.append(p1)
        p2 = p / ".." / ".." / "internal-examples" / "designs" / board_path
        if p2.is_dir():
            design_dirs.append(p2)

    files = None
    found_design_files = False
    for design_dir in design_dirs:
        files = _get_design_files_in_dir(design_dir)
        if files.programming_file:
            found_design_files = True
            break
    if not found_design_files:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), board_path)
    return files



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from chipscopy.tcf.services import Service

NAME = "Example"
"""Example service name."""


class ExampleProxy(Service):
    """TCF Example service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def __init__(self, channel):
        super(ExampleProxy, self).__init__(channel)
        self.listeners = {}

    def add_numbers(self, a, b, done):
        return self.send_xicom_command("addNumbers", (a, b), done)

    def parse_bin(self, device_type, data, done):
        params = {"device_type": device_type}
        return self.send_xicom_command("parseBin", (params, data), done)

    def mrd(self, node_id, addr, word_count, done):
        return self.send_xicom_command("mrd", (node_id, addr, word_count), done)

    def mwr(self, node_id, addr, words, word_count, done):
        return self.send_xicom_command("mwr", (node_id, addr, words, word_count), done)

    def run_vio_test(self, node_id, done):
        args = {"node_id": node_id}
        return self.send_xicom_command("runVIOTest", args, done)

    def run_mila_test(self, node_id, done):
        return self.send_xicom_command("runMILATest", (node_id,), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from math import ceil, sqrt, pow

# section Matplot lib
import matplotlib
from matplotlib.ticker import FixedLocator, FixedFormatter
from matplotlib.widgets import Button, RadioButtons

from chipscopy.dm.harden.noc_perfmon.noc_types import ddrmc_main_typedef
from chipscopy.api.noc.noc_perfmon_utils import (
    PerfTGController,
    NoCElement,
    get_noc_typedef_from_name,
)

# use this for Qt5
matplotlib.use("Qt5Agg")
# this if for kivy
# matplotlib.use('module://kivy.garden.matplotlib.backend_kivy')
# matplotlib.use('kivy.garden.matplotlib.bickend_kivy')
import matplotlib.pyplot as plt
import matplotlib as mpl

MIN_Y = 100


def is_row_start(col_count, index):
    # given col counts and an index return if it's a row start
    # 1-index is required
    if (index - 1) % col_count == 0:
        return True
    else:
        return False


def build_y_scale_labels(y_max, len_ay):
    # return the y-axis labels array
    scale = ""
    labels = ["0"]
    locations = [0]
    fm = mpl.ticker.EngFormatter(places=2)
    r = 1
    if 0 <= y_max < pow(1000, 1):
        # no scaling
        pass
    elif pow(1000, 1) <= y_max < pow(1000, 2):
        scale = "K"
        r = pow(1000, 1)
    elif pow(1000, 2) <= y_max < pow(1000, 3):
        scale = "M"
        r = pow(1000, 2)
    elif pow(1000, 3) <= y_max < pow(1000, 4):
        scale = "G"
        r = pow(1000, 3)
    elif pow(1000, 4) <= y_max < pow(1000, 5):
        scale = "T"
        r = pow(1000, 4)
    for x in range(1, len_ay):
        val = x / (len_ay - 1) * y_max
        # wradix = val/r
        wradix = fm(val)
        if x == 1:
            labels.insert(0, f"-{wradix}")
            locations.insert(0, -val)
        labels.append(f"{wradix}")
        locations.append(val)
    return labels, locations


class ViewButtonProcessor:
    def __init__(self, axes, label, plot):
        self.plot = plot
        self.label = label
        self.button = Button(axes, label)
        self.button.on_clicked(self.switch_view)

    def switch_view(self, event):
        self.plot.switch_view(self.label)


class MeasurementPlot:
    def __init__(
        self,
        enable_nodes,
        mock,
        mock_file="slice.json",
        figsize=None,
        kivy=False,
        tg: PerfTGController = None,
    ):
        self.enable_nodes = [x.lower() for x in enable_nodes]
        self.mock = mock
        self.mock_file = mock_file
        self.figsize = figsize
        self.kivy = kivy
        self.tg = tg
        self.first_render = True
        self.display_nodes = []  # because some views don't apply to all nodes
        self.alive = True
        self.switching = False
        self.errors_present = False

        self.listener = None
        self.ifh = None
        self.lineno = 0
        self.fig = None
        self.plots = {}
        self.backgrounds = {}
        self.subplots = {}
        self.error_text_boxes = {}
        self.refresh_count = 0
        self.num_elements = None
        self.mock_data = {}
        self.view = "bandwidth"
        self.previous = None
        self.pattern = "Write Linear"

        # matplotlib GC is zealous, hence:
        self.latency_button = None
        self.bandwidth_button = None
        self.view_buttons = []
        self.start_btn = None
        self.stop_btn = None
        self.back_btn = None
        self.pattern_group = None
        self.tg_setup_btn = None
        self.legend_line_map = {}
        self.units_radio = None

        # special handling for DDRMC
        self.views = ["bandwidth", "latency"]
        for node in self.enable_nodes:
            if get_noc_typedef_from_name(node) == ddrmc_main_typedef:
                self.views.append("ddrmc")
                break

    def get_incremental_data(self):
        if not self.mock:
            return self.listener.unique_elements, None
        else:
            if self.ifh is None:
                self.ifh = open(self.mock_file, "r")
            self.lineno += 1
            line = self.ifh.readline()
            temp = json.loads(line)
            self.mock_data.update(temp)
            noc_element_data = self.mock_data
            return noc_element_data, next(iter(temp.keys()))

    def build_graphs(self):
        if self.view == "bandwidth":
            unit = "B/s"
            bottom = 0.11
        elif self.view == "latency":
            self.units = "ns"
            bottom = 0.175
        elif self.view == "ddrmc":
            unit = "counts"
            bottom = 0.175

        # special handling for DDRMC
        if self.view == "ddrmc":
            self.display_nodes = []
            for node in self.enable_nodes:
                if get_noc_typedef_from_name(node) == ddrmc_main_typedef:
                    self.display_nodes.append(node)
        else:
            self.display_nodes = self.enable_nodes

        noc_element_data, _ = self.get_incremental_data()
        col_count = ceil(sqrt(len(self.display_nodes)))
        row_count = ceil(len(self.display_nodes) / col_count)
        if self.figsize is not None:
            self.fig = plt.figure(figsize=self.figsize)
        else:
            self.fig = plt.figure()

        plt.subplots_adjust(left=0.1, right=0.965, bottom=bottom, top=0.88, wspace=0.2, hspace=0.2)
        # plt.subplots_adjust(left=.05, right=.965, bottom=bottom, top=.88, wspace=.25, hspace=.2)

        # turn on interactive
        plt.ion()

        plt.get_current_fig_manager().set_window_title(f"DDR & NoC PerfMon {self.view} Plot")

        # TODO fixup
        all_found = False
        if self.mock:
            while not all_found:
                noc_element_data, _ = self.get_incremental_data()
                all_found = set(self.enable_nodes).issubset(set(noc_element_data.keys()))

        # unmap existing legend lines
        self.legend_line_map = {}
        # hook up handlers
        self.fig.canvas.mpl_connect("pick_event", self.on_pick)
        self.fig.canvas.mpl_connect("close_event", self.on_close)

        # special processing for latency view:
        if self.view == "latency":
            radio_ax = plt.axes([0.01, 0.01, 0.15, 0.15])
            self.units_radio = RadioButtons(radio_ax, ("ns", "clocks"))

            def latency_units(label):
                self.units = label

            self.units_radio.on_clicked(latency_units)

        plot_index = 0
        # for elem, storage in noc_element_data.items():
        for elem in self.display_nodes:
            storage = noc_element_data[elem]
            data = storage.samples
            plot_index += 1
            if plot_index == 1:
                self.num_elements = len(next(iter(data.values())))
            sp = plt.subplot(row_count, col_count, plot_index)
            self.plots[elem] = {"left_axis": {}, "right_axis": {}}
            x = list(range(50))
            for metric in storage.axis_metrics["left_axis"][self.view]:
                if self.view == "latency":
                    if self.units != "clocks":
                        # need to convert
                        tslide = noc_element_data[elem].tslide
                        counter_freq_mhz = noc_element_data[elem].counter_freq_mhz
                        converted_data = [
                            element * pow(2, tslide) * 1000 / counter_freq_mhz
                            for element in data[metric]
                        ]
                        lambda_plot = plt.plot(x, converted_data)[0]
                else:
                    lambda_plot = plt.plot(x, data[metric])[0]
                self.plots[elem]["left_axis"][metric] = lambda_plot

            lambda_plot = next(iter(self.plots[elem]["left_axis"].values()))
            lambda_plot.axes.set_xticks(range(0, self.num_elements)[::5])
            lambda_plot.axes.set_xticklabels([])

            # right axis metrics
            for metric in storage.axis_metrics["right_axis"][self.view]:
                right_lambda_axes = lambda_plot.axes.twinx()
                r_lambda_plot = right_lambda_axes.plot(x, data[metric])[0]
                self.plots[elem]["right_axis"][metric] = r_lambda_plot

            if is_row_start(col_count, plot_index):
                if self.view == "latency":
                    sp.set_ylabel(f"{self.view}")
                else:
                    sp.set_ylabel(f"{self.view} - {unit}")

            for axis in ["left_axis", "right_axis"]:
                if len(self.plots[elem][axis].keys()) == 0:
                    continue
                lambda_plot = next(iter(self.plots[elem][axis].values()))
                data_max = max([max(data[x]) for x in storage.axis_metrics[axis][self.view]])
                new_y_max = max(data_max, MIN_Y)
                new_y_min = -0.1 * new_y_max
                new_y_max = 1.1 * new_y_max
                lambda_plot.axes.set_ylim((new_y_min, new_y_max))

            sp.set_title(f"{elem}")
            self.backgrounds[elem] = self.fig.canvas.copy_from_bbox(lambda_plot.axes.bbox)
            self.subplots[elem] = sp
            self.error_text_boxes[elem] = {}
            # legend
            legend_keys = []
            for axis in ["left_axis", "right_axis"]:
                for key in storage.axis_metrics[axis][self.view]:
                    legend_keys.append(key)
            leg = plt.legend(legend_keys)

            # hook up pickers for visibility toggling
            # TODO: right axis
            for legend_line, line_plot in zip(
                leg.get_lines(), self.plots[elem]["left_axis"].values()
            ):
                # enable p
                legend_line.set_picker(True)
                legend_line.set_pickradius(5)
                self.legend_line_map[legend_line] = line_plot

        # navigation buttons buttons
        offset = 0
        self.view_buttons = []
        for view in self.views:
            if view == self.view:
                continue
            # [left, bottom, width, height]
            button_axes = plt.axes([0.8 + offset, 0.025, 0.075, 0.05])
            offset += 0.1
            self.view_buttons.append(ViewButtonProcessor(button_axes, view, self))

        # TG setup page
        if self.tg is not None:
            # [left, bottom, width, height]
            tg_setup_ax = plt.axes([0.6, 0.025, 0.075, 0.05])
            self.tg_setup_btn = Button(tg_setup_ax, "TG Setup")
            self.tg_setup_btn.on_clicked(self.switch_to_tg_setup)

        # if not self.kivy:
        #     # plt.show()
        #     pass

        if self.first_render:
            plt.show()
            self.first_render = False

        self.fig.canvas.draw()

    # noinspection PyUnboundLocalVariable
    def update(self, node: NoCElement):
        if self.view == "tg_control":
            return

        # TODO mock-data?
        node_to_update = node
        noc_element_data, updated = self.get_incremental_data()
        if self.mock:
            node_to_update = updated

        elem = node.name
        data = node.samples

        self.fig.canvas.restore_region(self.backgrounds[elem])
        for metric, plot in self.plots[elem]["left_axis"].items():
            if self.view == "latency":
                if self.units != "clocks":
                    # need to convert
                    tslide = node.tslide
                    counter_freq_mhz = node.counter_freq_mhz
                    converted_data = [sample * 1000 / counter_freq_mhz for sample in data[metric]]
                    plot.set_ydata(converted_data)
                else:
                    plot.set_ydata(data[metric])
            else:
                plot.set_ydata(data[metric])

        for metric, plot in self.plots[elem]["right_axis"].items():
            plot.set_ydata(data[metric])

        for axis in ["left_axis", "right_axis"]:
            if len(self.plots[elem][axis].keys()) == 0:
                continue
            lambda_plot = next(iter(self.plots[elem][axis].values()))
            data_max = 0
            for metric in self.plots[elem][axis].keys():
                for line in self.plots[elem][axis][metric].axes.lines:
                    line_max = max(line.get_ydata())
                    if line_max > data_max:
                        data_max = line_max
            # data_max = max([max(data[x]) for x in self.plots[elem][axis][:]])
            new_y_max = max(data_max, MIN_Y)
            new_y_min = -0.1 * new_y_max
            new_y_max = 1.1 * new_y_max
            lambda_plot.axes.set_ylim((new_y_min, new_y_max))
            y_labels, y_locations = build_y_scale_labels(max(data_max, MIN_Y), 6)
            lambda_plot.axes.yaxis.set_major_locator(FixedLocator(y_locations))
            lambda_plot.axes.yaxis.set_major_formatter(FixedFormatter(y_labels))

        # flags - search for any flags in any channel in any visible sample
        error_flags = []
        for sample in data["flags"]:
            if len(sample) > 0:
                for error in sample:
                    if error not in error_flags:
                        error_flags.append(error)

        error_y = 0.1
        for error in error_flags:
            if error not in self.error_text_boxes[elem].keys():
                self.error_text_boxes[elem][error] = self.subplots[elem].text(
                    0.1,
                    error_y,
                    error,
                    size=10,
                    ha="left",
                    va="top",
                    bbox=dict(boxstyle="square", ec=(1.0, 0.5, 0.5), fc=(1.0, 0.8, 0.8)),
                    transform=self.subplots[elem].transAxes,
                )
            else:
                self.error_text_boxes[elem][error].set_visible(True)
            error_y += 0.1

        for error in self.error_text_boxes[elem].keys():
            if error not in error_flags:
                self.error_text_boxes[elem][error].set_visible(False)

        # not going to do dynamic xtick labels
        # plot.axes.set_xticks(range(0, self.num_elements), data['ts'])
        # plot.axes.set_xticklabels(data['ts'][::5], rotation=90)
        # self.fig.canvas.blit(plot.axes.clipbox)

        # Attempts to speed up plotting
        # plot.axes.draw_artist(plot.axes.patch)

        # failed to update xtick labels
        # plot.axes.draw_artist(plot.axes.get_xticks())

        # working in standalone app
        # plot.axes.draw_artist(plot)
        # self.fig.canvas.blit(plot.axes.bbox)

        # notebook trials
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

        # self.fig.canvas.draw_idle() -- this didn't work

        # self.refresh_count += 1
        # if self.refresh_count % 100 == 0:
        #     self.fig.canvas.draw()
        #     self.refresh_count = 0

    def build_tg_controlboard(self):
        if self.figsize is not None:
            self.fig = plt.figure(figsize=self.figsize)
        else:
            self.fig = plt.figure()

        # turn on interactive
        plt.ion()

        # set the title
        plt.get_current_fig_manager().set_window_title(
            "DDR & NoC PerfMon Traffic Generator Control"
        )

        def on_pattern_select(label):
            self.pattern = label

        # [left, bottom, width, height]
        pattern_ax = plt.axes([0.4, 0.5, 0.2, 0.1])
        patterns = tuple(list(self.tg.supported_patterns))
        active_pattern_index = list(self.tg.supported_patterns).index(self.pattern)
        self.pattern_group = RadioButtons(pattern_ax, patterns, active=active_pattern_index)

        self.pattern_group.on_clicked(on_pattern_select)

        def on_start(event):
            self.tg.set_active_pattern(self.pattern)
            self.tg.program()
            self.tg.start()

        # [left, bottom, width, height]
        start_tg_ax = plt.axes([0.4, 0.025, 0.075, 0.05])
        self.start_btn = Button(start_tg_ax, "Start TG")
        self.start_btn.on_clicked(on_start)

        def on_stop(event):
            self.tg.stop()

        # [left, bottom, width, height]
        stop_tg_ax = plt.axes([0.5, 0.025, 0.075, 0.05])
        self.stop_btn = Button(stop_tg_ax, "Stop TG")
        self.stop_btn.on_clicked(on_stop)

        def on_back(event):
            # if self.previous == 'bandwidth':
            #     self.switch_to_bandwidth(event=event)
            # elif self.previous == 'latency':
            #     self.switch_to_latency(event=event)
            self.switch_view(self.previous)

        # [left, bottom, width, height]
        back_ax = plt.axes([0.6, 0.025, 0.075, 0.05])
        self.back_btn = Button(back_ax, "Back")
        self.back_btn.on_clicked(on_back)

        # if not self.kivy:
        #     # plt.show()
        #     pass

        self.fig.canvas.draw()

    def on_pick(self, event):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legend_key = event.artist
        line_plot = self.legend_line_map[legend_key]
        vis = not line_plot.get_visible()
        line_plot.set_visible(vis)
        # Change the alpha on the line in the legend so we can see what lines
        # have been toggled
        if vis:
            legend_key.set_alpha(1.0)
        else:
            legend_key.set_alpha(0.2)
        self.fig.canvas.draw()

    def on_close(self, evt):
        if not self.switching:
            self.alive = False

    def destroy(self):
        self.switching = True
        plt.close(self.fig)
        self.switching = False

    def switch_view(self, view):
        self.previous = self.view
        self.view = view
        self.destroy()
        self.build_graphs()

    def switch_to_latency(self, event):
        self.previous = self.view
        self.view = "latency"
        self.destroy()
        # self.build_latency_graphs()
        self.build_graphs()

    def switch_to_bandwidth(self, event):
        self.previous = self.view
        self.view = "bandwidth"
        self.destroy()
        # self.build_bandwidth_graphs()
        self.build_graphs()

    def switch_to_tg_setup(self, event):
        self.previous = self.view
        self.view = "tg_control"
        self.destroy()
        self.build_tg_controlboard()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DebugCorePolling service provides operations to manage polling for DebugCore nodes in the Hardware Server.

.. |DebugCorePoll| replace:: :class:`DebugCorePoll`

General Flow:

1. Searches through list of polls and re-uses if desired
2. Register polling DebugCore sequence with hw_server
3. hw_server runs registered polls periodically
    a) performs some matching (determining whether some condition is hit)
    b) sends events when condition is met
4. Deregister poll(s) with hw_server

DebugCoreSequence
^^^^^^^^^^^^^^^^^
List of property |dict| describing a transaction sequence

+-------------------+----------------------+------------------------------------------+
| Name              | Type                 | Description                              |
+===================+======================+==========================================+
| type              | |str|                | Transaction type                         |
|                   |                      |   Read: 'r'                              |
|                   |                      |     Reads from an address                |
|                   |                      |     and returns data                     |
|                   |                      |   Write: 'w'                             |
|                   |                      |     Writes to an address                 |
|                   |                      |   Packet: 'p'                            |
|                   |                      |     Runs a data packet on a node and     |
|                   |                      |     returns result data                  |
+-------------------+----------------------+------------------------------------------+
| name              | |str|                | Name of the seq given by the client      |
+-------------------+----------------------+------------------------------------------+
| addr              | |int|                | Starting byte address of transaction     |
+-------------------+----------------------+------------------------------------------+
| data              | |bytearray|          | Data buffer for transaction.  Used for   |
|                   |                      | both writing and reading buffers         |
+-------------------+----------------------+------------------------------------------+
| size              | |int|                | Number of bytes to read during Read      |
|                   |                      | (default size=4)                         |
|                   |                      | (not used for Write and Packet types)    |
+-------------------+----------------------+------------------------------------------+
| read_size         | |int|                | Number of bytes to read during Packet    |
|                   |                      | transaction.  If not set or zero, don't  |
|                   |                      | read back.                               |
+-------------------+----------------------+------------------------------------------+

Examples
--------
Read/Write/Read Sequence:

::

    seq = [{'type':'r','addr':0x100,'size':4},{'type':'w',addr:0x100,'data':b'ABCD'},{'type':'r','addr':0x100,'size':4}]

    result = [{'type':'r','addr':0x100,'data':b'\x00\x00\x00\x00'},{'type':'r','addr':0x100,'data':b'ABCD'}]


DebugCorePoll
^^^^^^^^^^^^^
Dictionary of properties for a poll

+-------------------+----------------------+------------------------------------------+
| Name              | Type                 | Description                              |
+===================+======================+==========================================+
| name              | |str|                | Name of the Poll given by the client     |
+-------------------+----------------------+------------------------------------------+
| seq               | `DebugCoreSequence`_ | DebugCore Sequence to run at each poll   |
+-------------------+----------------------+------------------------------------------+
| period            | |int|                | Minimum time between polls (ms)          |
+-------------------+----------------------+------------------------------------------+

Extension of `DebugCoreSequence`_ for polling
---------------------------------------------

+-------------------+----------------------+------------------------------------------+
| Name              | Type                 | Description                              |
+===================+======================+==========================================+
| cond              | |str|                | Matching condition string used for       |
|                   |                      | determining a triggering condition.      |
|                   |                      | `Condition Syntax`_                      |
+-------------------+----------------------+------------------------------------------+
| cond_offset       | |int|                | Bit offset of condition matching.        |
|                   |                      | For 'p' type default is 32bits, 0 for 'r'|
+-------------------+----------------------+------------------------------------------+
| bitwise_op        | |str|                | Bitwise operation:                       |
|                   |                      |   'and':                                 |
|                   |                      |     all bits must meet conditions        |
|                   |                      |   'or':                                  |
|                   |                      |     any bit meets condition (default)    |
+-------------------+----------------------+------------------------------------------+

Condition Syntax
----------------
A string indicating how each bit of transaction result is to be matched, low-endian. If any bit in the string meets
the condition then the whole poll is considered triggered.  The remaining transactions are completed and then the
results of the sequence are sent as an event to the listeners of the poll.

+-------------+----------------------------------+
|'X'          | don't care                       |
+-------------+----------------------------------+
|'0'          | match on 0                       |
+-------------+----------------------------------+
|'1'          | match on 1                       |
+-------------+----------------------------------+
|'T'          | toggled value                    |
+-------------+----------------------------------+
|'R'          | rising edge - 0 to 1             |
+-------------+----------------------------------+
|'F'          | falling edge - 1 to 0            |
+-------------+----------------------------------+

Short hand supported strings:

+-------------+----------------------------------+
|'always'     | always throw an event            |
+-------------+----------------------------------+
|'any_one'    | check is any bit is set          |
+-------------+----------------------------------+
|'changes'    | check for a change in any bit    |
+-------------+----------------------------------+

Example:

::

    Check only bit 4 for a changed value = "XXXXTXXXXXXXXXXXXXXXXXXXXXXXXXXX"

    Check if any bit is a changed value = "TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT" or "changes"


Service Interface
^^^^^^^^^^^^^^^^^
.. autoclass:: DebugCorePollingProxy
    :members: get_polls, get_context, add_listener, remove_listener, add_poll, remove_poll, retain_poll


Service Events
^^^^^^^^^^^^^^
.. autoclass:: DebugCorePollingListener
    :members:

"""

from typing import Dict, Any, NewType
from chipscopy.tcf.channel import EventListener, fromJSONSequence
from chipscopy.tcf.services import Service, DoneHWCommand, Token
from chipscopy.tcf.services.arguments import from_xargs


NAME = "DebugCorePolling"
"""DebugCorePolling service name."""

DebugCorePoll = NewType("DebugCorePoll", Dict[str, Any])

DebugCoreSequence = NewType("DebugCoreSequence", Dict[str, Any])


class DebugCorePollingListener(object):
    """Debug node event listener is notified when a poll for a DebugNode has been triggered."""

    def poll_added(self, poll: DebugCorePoll):
        """
        Called when a new poll has been registered

        :param poll: Properties of poll
        """
        pass

    def poll_removed(self, poll_ctx: str):
        """
        Called when a poll has been removed.

        :param poll_ctx: Poll context id
        """
        pass

    def poll_event(self, poll_ctx: str, props: Dict):
        """
        Called a registered poll condition has been triggered.

        :param poll_ctx: Poll context id
        :param props: Result sequence and other properties of triggering poll
        """
        pass


class DebugCorePollingProxy(Service):
    """TCF DebugCorePolling service interface."""

    def __init__(self, channel):
        super(DebugCorePollingProxy, self).__init__(channel)
        self.listeners = {}

    def getName(self):
        return NAME

    def get_polls(self, node_id: str = "", done: DoneHWCommand = None) -> Token:
        """
        Gets list of polls registered for the given node ("" means give all polls).

        :param node_id: Context Node ID
        :param done: Callback when command is complete
        :return: Token of request
        :done result: List of `DebugCorePoll`_
        """
        return self.send_xicom_command("getPolls", (node_id,), done)

    def get_context(self, poll_id: str, done: DoneHWCommand) -> Token:
        """
        Gets context properties of a poll.

        :param poll_id: Context Poll ID
        :param done: Callback when command is complete
        :return: Token of request
        """
        return self.send_xicom_command("getContext", (poll_id,), done)

    def add_poll(self, node_id: str, poll: DebugCorePoll, done: DoneHWCommand = None) -> Token:
        """
        Registers a new `DebugCorePoll`_ with the Hardware Server.

        :param node_id: Context node ID
        :param poll: Dict describing the poll
        :param done: Callback when command is complete
        :returns: Token of command request
        :done result: Context ID of new `DebugCorePoll`_
        """
        return self.send_xicom_command("addPoll", (node_id, poll), done)

    def remove_poll(self, poll_id: str, done: DoneHWCommand = None) -> Token:
        """
        Deregisters a `DebugCorePoll`_ with the Hardware Server.

        :param poll_id: Context poll ID
        :param done: Callback when command is complete
        :returns: Token of command request
        """
        return self.send_xicom_command("removePoll", (poll_id,), done)

    def retain_poll(self, poll_id: str, done: DoneHWCommand = None) -> Token:
        """
        Keeps a poll even if the original requestor of the poll removes it.  This also adds this peer to the list of
        listeners of the poll from the Hardware Server.

        :param poll_id: Context poll ID
        :param done: Callback when command is complete
        :returns: Token of command request
        """
        return self.send_xicom_command("retainPoll", (poll_id,), done)

    def add_listener(self, listener: DebugCorePollingListener):
        """Add DebugCore service event listener.

        :param listener: Event listener implementation.
        """
        channel_listener = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, channel_listener)
        self.listeners[listener] = channel_listener

    def remove_listener(self, listener: DebugCorePollingListener):
        """Remove DebugCore service event listener.

        :param listener: Event listener implementation.
        """
        channel_listener = self.listeners.get(listener)
        if channel_listener:
            del self.listeners[listener]
            self.channel.removeEventListener(self, channel_listener)


class ChannelEventListener(EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener
        self._event_handlers = {
            "pollAdded": self._poll_add,
            "pollRemoved": self._poll_remove,
            "pollEvent": self._poll_event,
        }

    def _poll_add(self, args):
        args = from_xargs(args)
        assert len(args) == 1
        poll = args[0]
        self.listener.poll_added(poll)

    def _poll_remove(self, args):
        args = from_xargs(args)
        assert len(args) == 1
        poll_id = args[0]
        self.listener.poll_removed(poll_id)

    def _poll_event(self, args):
        args = from_xargs(args)
        assert len(args) == 2
        poll_id = args[0]
        props = args[1]
        self.listener.poll_event(poll_id, props)

    def event(self, name, data):
        try:
            args = fromJSONSequence(data)
            try:
                handler = self._event_handlers[name]
                handler(args)
            except KeyError:
                raise IOError("DebugCorePolling service: unknown event: " + name)
        except Exception as x:
            import sys

            x.tb = sys.exc_info()[2]
            self.service.channel.terminate(x)


DebugCorePollingService = DebugCorePollingProxy
